# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for assembling signed source packages."""

import datetime as dt
import json
import logging
import re
import shutil
import subprocess
import tarfile
from pathlib import Path, PurePath
from textwrap import dedent
from typing import Any, AnyStr, override
from unittest import mock

from debian import deb822

import debusine.worker.tags as wtags
from debusine.artifacts import LocalArtifact, SourcePackage, Upload
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    DebusineSigningInput,
    DebusineSigningOutput,
    WorkRequestResults,
)
from debusine.artifacts.playground import ArtifactPlayground
from debusine.assets import KeyPurpose
from debusine.client.models import (
    ArtifactResponse,
    RelationsResponse,
    RemoteArtifact,
)
from debusine.tasks import AssembleSignedSource, FALLBACK_ARCHITECTURE
from debusine.tasks.assemble_signed_source import AssembleError
from debusine.tasks.models import (
    AssembleSignedSourceDynamicData,
    BaseTaskInputArtifact,
    InputArtifactMultiple,
    WorkerType,
)
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_artifact_response,
    create_file_response,
    create_remote_artifact,
    create_system_tarball_data,
)


class AssembleSignedSourceTests(
    ExternalTaskHelperMixin[AssembleSignedSource], TestCase
):
    """Tests for AssembleSignedSource."""

    SAMPLE_TASK_DATA = {
        "environment": "debian/match:codename=bookworm",
        "template": 11,
        # TODO: We should test multiple signed artifacts once more than one
        # KeyPurpose is defined.
        "signed": [12],
    }

    @override
    def setUp(self) -> None:
        super().setUp()

        self.resolver.set_field_artifact(
            "environment",
            id=1,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )
        self.resolver.set_field_artifact(
            "template",
            id=11,
            category=ArtifactCategory.BINARY_PACKAGE,
            data=DebianBinaryPackage(
                srcpkg_name="template",
                srcpkg_version="1.0-1",
                deb_fields={
                    "Package": "hello",
                    "Version": "1.0-1",
                    "Architecture": "amd64",
                },
                deb_control_files=[],
            ),
        )
        self.resolver.set_field_artifacts(
            "signed",
            BaseTaskInputArtifact(
                artifact_id=12,
                category=ArtifactCategory.SIGNING_OUTPUT,
                data=DebusineSigningOutput(
                    purpose=KeyPurpose.UEFI, fingerprint="", results=[]
                ),
            ),
        )

        self.resolver.field_values["unsigned"] = InputArtifactMultiple(
            label="label",
            lookup=["test"],
            artifacts=[
                BaseTaskInputArtifact(
                    artifact_id=13,
                    category=ArtifactCategory.SIGNING_INPUT,
                    data=DebusineSigningInput(),
                ),
                BaseTaskInputArtifact(
                    artifact_id=14,
                    category=ArtifactCategory.BINARY_PACKAGE,
                    data=DebianBinaryPackage(
                        srcpkg_name="unsigned",
                        srcpkg_version="1.0-1",
                        deb_fields={
                            "Package": "hello-unsigned",
                            "Version": "1.0-1",
                            "Architecture": "amd64",
                        },
                        deb_control_files=[],
                    ),
                ),
            ],
        )

    def test_can_run_on(self) -> None:
        """can_run_on returns True if unshare is available."""
        self.assertTrue(
            self.can_run_on(
                {
                    "system:worker_type": WorkerType.EXTERNAL,
                    "executor:unshare:available": True,
                    "assemblesignedsource:version": self.task.TASK_VERSION,
                    "system:architectures": [FALLBACK_ARCHITECTURE],
                }
            )
        )

    def test_can_run_on_mismatched_task_version(self) -> None:
        """can_run_on returns False for mismatched task versions."""
        self.assertFalse(
            self.can_run_on(
                {
                    "system:worker_type": WorkerType.EXTERNAL,
                    "executor:unshare:available": True,
                    "assemblesignedsource:version": self.task.TASK_VERSION + 1,
                }
            )
        )

    def test_can_run_on_unshare_not_available(self) -> None:
        """can_run_on returns False if unshare is not available."""
        self.assertFalse(
            self.can_run_on(
                {
                    "system:worker_type": WorkerType.EXTERNAL,
                    "executor:unshare:available": False,
                    "assemblesignedsource:version": self.task.TASK_VERSION,
                }
            )
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            require_system={
                wtags.BUILD_ARCH_PREFIX + "amd64",
                wtags.EXECUTOR_PREFIX + "unshare",
            }
        )

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            AssembleSignedSourceDynamicData(
                environment_id=1,
                template_id=11,
                signed_ids=[12],
                unsigned_ids=[13, 14],
                subject="hello",
                parameter_summary="hello_1.0-1 (amd64)",
                runtime_context=None,
            ),
        )

    def test_get_subject(self) -> None:
        self.assertEqual(self.get_subject(), "hello")

    def test_get_configuration_context(self) -> None:
        self.assertIsNone(self.get_configuration_context())

    def test_acceptable_input_categories(self) -> None:
        self.assertEqual(
            AssembleSignedSource.inputs["template"].categories,  # type: ignore[attr-defined]
            (ArtifactCategory.BINARY_PACKAGE,),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

        self.task.dynamic_data = AssembleSignedSourceDynamicData(
            environment_id=1,
            template_id=2,
            signed_ids=[3],
            unsigned_ids=[4],
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1, 2, 3, 4])

    def test_fetch_input_template_wrong_category(self) -> None:
        """fetch_input checks the category of the template artifact."""
        self.task.dynamic_data = AssembleSignedSourceDynamicData(
            environment_id=1, template_id=11, signed_ids=[12]
        )
        debusine_mock = self.mock_debusine()
        debusine_mock.download_artifact.return_value = create_artifact_response(
            id=11, category=ArtifactCategory.TEST
        )

        self.assertFalse(
            self.task.fetch_input(self.create_temporary_directory())
        )

        self.assertEqual(
            (self.task.log_directory / "fetch_input.log").read_text(),
            "Expected template of category debian:binary-package; got "
            "debusine:test\n",
        )

    def test_fetch_input_template_no_files(self) -> None:
        """fetch_input fails if the template has no .deb files."""
        self.task.dynamic_data = AssembleSignedSourceDynamicData(
            environment_id=1, template_id=11, signed_ids=[12]
        )
        debusine_mock = self.mock_debusine()
        debusine_mock.download_artifact.return_value = create_artifact_response(
            id=11,
            category=ArtifactCategory.BINARY_PACKAGE,
            files={"foo": create_file_response()},
        )

        self.assertFalse(
            self.task.fetch_input(self.create_temporary_directory())
        )

        self.assertEqual(
            (self.task.log_directory / "fetch_input.log").read_text(),
            "Expected exactly one .deb package in template; got []\n",
        )

    def test_fetch_input_template_multiple_deb_files(self) -> None:
        """fetch_input fails if the template has multiple .deb files."""
        self.task.dynamic_data = AssembleSignedSourceDynamicData(
            environment_id=1, template_id=11, signed_ids=[12]
        )
        debusine_mock = self.mock_debusine()
        debusine_mock.download_artifact.return_value = create_artifact_response(
            id=11,
            category=ArtifactCategory.BINARY_PACKAGE,
            files={
                "a.deb": create_file_response(),
                "b.deb": create_file_response(),
            },
        )

        self.assertFalse(
            self.task.fetch_input(self.create_temporary_directory())
        )

        self.assertEqual(
            (self.task.log_directory / "fetch_input.log").read_text(),
            "Expected exactly one .deb package in template; got "
            "['a.deb', 'b.deb']\n",
        )

    def _create_binary_package_response(
        self, artifact_id: int
    ) -> ArtifactResponse:
        return create_artifact_response(
            id=artifact_id,
            category=ArtifactCategory.BINARY_PACKAGE,
            data={
                "srcpkg_name": "hello-signed",
                "srcpkg_version": "1.0+1",
                "deb_fields": {"Package": "hello-signed"},
                "deb_control_files": [],
            },
            files={"hello-signed_1.0+1_amd64.deb": create_file_response()},
        )

    def test_fetch_input_signed_wrong_category(self) -> None:
        """fetch_input checks the category of signed artifacts."""
        self.task.dynamic_data = AssembleSignedSourceDynamicData(
            environment_id=1, template_id=11, signed_ids=[12]
        )
        debusine_mock = self.mock_debusine()
        debusine_mock.download_artifact.side_effect = [
            self._create_binary_package_response(11),
            create_artifact_response(id=12, category=ArtifactCategory.TEST),
        ]

        self.assertFalse(
            self.task.fetch_input(self.create_temporary_directory())
        )

        self.assertEqual(
            (self.task.log_directory / "fetch_input.log").read_text(),
            "Expected template of category debusine:signing-output; got "
            "debusine:test\n",
        )

    def test_fetch_input_signed_gathers_files_and_built_using(self) -> None:
        """fetch_input gathers information from signed artifacts."""
        self.task.dynamic_data = AssembleSignedSourceDynamicData(
            environment_id=1, template_id=11, signed_ids=[12]
        )
        debusine_mock = self.mock_debusine()
        debusine_mock.download_artifact.side_effect = [
            self._create_binary_package_response(11),
            create_artifact_response(
                id=12,
                category=ArtifactCategory.SIGNING_OUTPUT,
                data={
                    "purpose": "uefi",
                    "fingerprint": "0" * 64,
                    "results": [
                        {
                            "file": "hello/hello-1.efi",
                            "output_file": "hello/hello-1.efi.sig",
                        },
                        {"file": "hello/hello-2.efi", "error_message": "Boom"},
                    ],
                },
            ),
        ]
        artifact_categories = {
            2: ArtifactCategory.TEST,
            3: ArtifactCategory.BINARY_PACKAGE,
            4: ArtifactCategory.SIGNING_INPUT,
        }
        relations = {
            4: [(2, "extends"), (3, "relates-to")],
            12: [(2, "relates-to"), (4, "relates-to")],
        }

        def relation_list(artifact_id: int) -> RelationsResponse:
            return RelationsResponse.model_validate(
                [
                    {
                        "id": 1,
                        "artifact": artifact_id,
                        "target": target_id,
                        "type": relation_type,
                    }
                    for target_id, relation_type in relations[artifact_id]
                ]
            )

        def artifact_get(
            artifact_id: int,
            workspace: str,  # noqa: ARG001, U100
        ) -> ArtifactResponse:
            return create_artifact_response(
                id=6, category=artifact_categories[artifact_id]
            )

        debusine_mock.relation_list.side_effect = relation_list
        debusine_mock.artifact_get.side_effect = artifact_get
        destination = self.create_temporary_directory()

        self.assertTrue(self.task.fetch_input(destination))

        self.assertEqual(
            self.task._signed_files,
            {
                (KeyPurpose.UEFI, "hello/hello-1.efi"): (
                    destination
                    / "signed-artifacts"
                    / "12"
                    / "hello"
                    / "hello-1.efi.sig"
                )
            },
        )
        self.assertEqual(self.task._built_using_ids, [3])

    def test_configure_for_execution(self) -> None:
        """configure_for_execution installs dependencies."""
        download_directory = self.create_temporary_directory()
        mocked_prepare_executor_instance = (
            self.patch_prepare_executor_instance()
        )

        self.assertTrue(self.task.configure_for_execution(download_directory))

        mocked_prepare_executor_instance.assert_called_once_with()
        assert isinstance(self.task.executor_instance, mock.MagicMock)
        self.task.executor_instance.assert_has_calls(
            [
                mock.call.run(
                    ["apt-get", "update"], run_as_root=True, check=True
                ),
                mock.call.run(
                    ["apt-get", "--yes", "install", "dpkg-dev"],
                    run_as_root=True,
                    check=True,
                ),
            ]
        )

    def test_pull_from_executor(self) -> None:
        """_pull_from_executor pulls a directory from the executor using tar."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        self.assertTrue(self.task.configure_for_execution(download_directory))
        assert isinstance(self.task.executor_instance, mock.MagicMock)
        self.task.executor_instance.run.reset_mock()
        self.task.executor_instance.run.return_value = (
            subprocess.CompletedProcess([], 0)
        )

        self.task._remote_execute_directory = self.create_temporary_directory()
        self.task._local_execute_directory = self.create_temporary_directory()
        source = self.task._remote_execute_directory / "directory"
        target = self.task._local_execute_directory / "directory"
        source_tar = self.task._remote_execute_directory / "directory.tar"
        target_tar = self.task._local_execute_directory / "directory.tar"

        with mock.patch("subprocess.run") as mock_run:
            self.task._pull_from_executor(source, target)

        self.task.executor_instance.run.assert_called_once_with(
            ["tar", "-cf", f"{source}.tar", "directory"],
            run_as_root=False,
            cwd=self.task._remote_execute_directory,
            env=self.task._cmd_env(),
            stdout=mock.ANY,
            stderr=mock.ANY,
        )
        self.task.executor_instance.file_pull.assert_called_once_with(
            source_tar, target_tar
        )
        mock_run.assert_called_once_with(
            ["tar", "--one-top-level", "-xf", target_tar],
            cwd=self.task._local_execute_directory,
        )

    def test_push_to_executor(self) -> None:
        """_push_to_executor pushes a directory to the executor using tar."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        self.assertTrue(self.task.configure_for_execution(download_directory))
        assert isinstance(self.task.executor_instance, mock.MagicMock)
        self.task.executor_instance.run.reset_mock()
        self.task.executor_instance.run.return_value = (
            subprocess.CompletedProcess([], 0)
        )

        self.task._local_execute_directory = self.create_temporary_directory()
        self.task._remote_execute_directory = self.create_temporary_directory()
        source = self.task._local_execute_directory / "directory"
        target = self.task._remote_execute_directory / "directory"
        source_tar = self.task._local_execute_directory / "directory.tar"
        target_tar = self.task._remote_execute_directory / "directory.tar"

        with mock.patch("subprocess.run") as mock_run:
            self.task._push_to_executor(source, target)

        mock_run.assert_called_once_with(
            ["tar", "-cf", source_tar, "directory"],
            cwd=self.task._local_execute_directory,
        )
        self.task.executor_instance.file_push.assert_called_once_with(
            source_tar, target_tar
        )
        self.task.executor_instance.run.assert_called_once_with(
            ["tar", "--one-top-level", "-xf", f"{target}.tar"],
            run_as_root=False,
            cwd=self.task._remote_execute_directory,
            env=self.task._cmd_env(),
            stdout=mock.ANY,
            stderr=mock.ANY,
        )

    def patch_executor_methods(self) -> tuple[mock.MagicMock, mock.MagicMock]:
        """
        Patch _pull_from_executor and _push_to_executor to do local copies.

        Returns (mock_pull_from_executor, mock_push_to_executor).
        """
        pull_from_executor_patcher = mock.patch.object(
            self.task, "_pull_from_executor", side_effect=shutil.copytree
        )
        mock_pull_from_executor = pull_from_executor_patcher.start()
        self.addCleanup(pull_from_executor_patcher.stop)

        push_to_executor_patcher = mock.patch.object(
            self.task, "_push_to_executor", side_effect=shutil.copytree
        )
        mock_push_to_executor = push_to_executor_patcher.start()
        self.addCleanup(push_to_executor_patcher.stop)

        return mock_pull_from_executor, mock_push_to_executor

    def test_extract_binary_error(self) -> None:
        """_extract_binary raises AssembleError if dpkg-deb exits non-zero."""
        execute_directory = self.create_temporary_directory()

        with (
            self.assertRaisesRegex(
                AssembleError, "dpkg-deb exited with code 1"
            ),
            mock.patch.object(self.task, "run_cmd", return_value=1),
        ):
            self.task._extract_binary(
                Path("foo.deb"), execute_directory / "template"
            )

    def test_extract_binary_success(self) -> None:
        """_extract_binary calls dpkg-deb properly."""
        execute_directory = self.create_temporary_directory()

        with (
            self.assertLogs("debusine.tasks", level=logging.INFO) as log,
            mock.patch.object(
                self.task, "run_cmd", return_value=0
            ) as mock_run_cmd,
        ):
            self.task._extract_binary(
                Path("foo.deb"), execute_directory / "template"
            )

        mock_run_cmd.assert_called_once_with(
            ["dpkg-deb", "-x", "foo.deb", str(execute_directory / "template")],
            execute_directory,
        )
        self.assertEqual(
            log.output,
            [
                f"INFO:debusine.tasks:Executing: dpkg-deb -x foo.deb "
                f"{execute_directory}/template",
                "INFO:debusine.tasks:dpkg-deb exited with code 0",
            ],
        )

    def _create_source_template(self, source_template_path: Path) -> None:
        format_path = source_template_path / "debian/source/format"
        format_path.parent.mkdir(parents=True)
        format_path.write_text("3.0 (native)\n")
        for file in (
            "debian/changelog",
            "debian/control",
            "debian/copyright",
            "debian/rules",
        ):
            (source_template_path / file).touch()

    def test_check_template_required_file(self) -> None:
        """_check_template checks that various files are present."""
        code_signing_path = self.create_temporary_directory()
        source_template_path = code_signing_path / "source-template"
        self._create_source_template(source_template_path)
        for file in (
            "debian/source/format",
            "debian/changelog",
            "debian/control",
            "debian/copyright",
            "debian/rules",
        ):
            (source_template_path / file).unlink()
            with (
                self.subTest(file=file),
                self.assertRaisesRegex(
                    AssembleError, f"Required file {file} missing from template"
                ),
            ):
                self.task._check_template(code_signing_path)
            (source_template_path / file).touch()

    def test_check_template_wrong_source_format(self) -> None:
        """_check_template requires a 3.0 (native) source format."""
        code_signing_path = self.create_temporary_directory()
        source_template_path = code_signing_path / "source-template"
        self._create_source_template(source_template_path)
        (source_template_path / "debian/source/format").write_text(
            "3.0 (quilt)\n"
        )
        with self.assertRaisesRegex(
            AssembleError,
            re.escape(
                "Expected template source format '3.0 (native)'; "
                "got '3.0 (quilt)'"
            ),
        ):
            self.task._check_template(code_signing_path)

    def test_check_template_forbidden_file(self) -> None:
        """_check_template checks that various files are absent."""
        code_signing_path = self.create_temporary_directory()
        source_template_path = code_signing_path / "source-template"
        self._create_source_template(source_template_path)
        for file in ("debian/source/options", "debian/source/local-options"):
            (source_template_path / file).touch()
            with (
                self.subTest(file=file),
                self.assertRaisesRegex(
                    AssembleError, f"Template may not contain {file}"
                ),
            ):
                self.task._check_template(code_signing_path)
            (source_template_path / file).unlink()

    def test_check_template_ok(self) -> None:
        """_check_template accepts a well-formed template."""
        code_signing_path = self.create_temporary_directory()
        source_template_path = code_signing_path / "source-template"
        self._create_source_template(source_template_path)
        self.task._check_template(code_signing_path)

    def test_add_signed_files_invalid_package_name(self) -> None:
        """_add_signed_files fails if a package name is invalid."""
        code_signing_path = self.create_temporary_directory()
        source_path = self.create_temporary_directory()
        (code_signing_path / "files.json").write_text(
            json.dumps({"packages": {"_invalid": {}}})
        )

        with self.assertRaisesRegex(
            AssembleError, "'_invalid' is not a valid package name"
        ):
            self.task._add_signed_files(code_signing_path, source_path)

    def test_add_signed_files_file_with_parent_segment(self) -> None:
        """_add_signed_files fails if a file name contains ".."."""
        code_signing_path = self.create_temporary_directory()
        source_path = self.create_temporary_directory()
        (code_signing_path / "files.json").write_text(
            json.dumps(
                {
                    "packages": {
                        "hello": {
                            "files": [
                                {
                                    "sig_type": "efi",
                                    "file": "usr/../../etc/passwd",
                                }
                            ]
                        }
                    }
                }
            )
        )

        with self.assertRaisesRegex(
            AssembleError,
            re.escape(
                "File name 'usr/../../etc/passwd' may not contain '..' segments"
            ),
        ):
            self.task._add_signed_files(code_signing_path, source_path)

    def test_add_signed_files_absolute_file(self) -> None:
        """_add_signed_files fails if a file name is absolute."""
        code_signing_path = self.create_temporary_directory()
        source_path = self.create_temporary_directory()
        (code_signing_path / "files.json").write_text(
            json.dumps(
                {
                    "packages": {
                        "hello": {
                            "files": [
                                {"sig_type": "efi", "file": "/boot/vmlinuz"}
                            ]
                        }
                    }
                }
            )
        )

        with self.assertRaisesRegex(
            AssembleError, "File name '/boot/vmlinuz' may not be absolute"
        ):
            self.task._add_signed_files(code_signing_path, source_path)

    def test_add_signed_files_unhandled_sig_type(self) -> None:
        """_add_signed_files fails if a file has an unknown sig_type."""
        code_signing_path = self.create_temporary_directory()
        source_path = self.create_temporary_directory()
        (code_signing_path / "files.json").write_text(
            json.dumps(
                {
                    "packages": {
                        "hello": {
                            "files": [
                                {"sig_type": "unknown", "file": "boot/vmlinuz"}
                            ]
                        }
                    }
                }
            )
        )

        with self.assertRaisesRegex(
            AssembleError, "Cannot handle sig_type 'unknown'"
        ):
            self.task._add_signed_files(code_signing_path, source_path)

    def test_add_signed_files_missing_signature(self) -> None:
        """_add_signed_files if a required signature is not available."""
        code_signing_path = self.create_temporary_directory()
        source_path = self.create_temporary_directory()
        (code_signing_path / "files.json").write_text(
            json.dumps(
                {
                    "packages": {
                        "hello": {
                            "files": [
                                {"sig_type": "efi", "file": "boot/vmlinuz"}
                            ]
                        }
                    }
                }
            )
        )

        with self.assertRaisesRegex(
            AssembleError,
            "uefi signature of 'hello/boot/vmlinuz' not available",
        ):
            self.task._add_signed_files(code_signing_path, source_path)

    def test_add_signed_files_copies_signatures(self) -> None:
        """_add_signed_files copies signatures into the source path."""
        download_path = self.create_temporary_directory()
        code_signing_path = self.create_temporary_directory()
        source_path = self.create_temporary_directory()
        (code_signing_path / "files.json").write_text(
            json.dumps(
                {
                    "packages": {
                        "hello": {
                            "files": [
                                {"sig_type": "efi", "file": "boot/vmlinuz"}
                            ]
                        }
                    }
                }
            )
        )
        signed_vmlinuz_path = (
            download_path / "signed-artifacts" / "12" / "boot" / "vmlinuz"
        )
        signed_vmlinuz_path.parent.mkdir(parents=True)
        signed_vmlinuz_path.write_text("signed vmlinuz\n")
        self.task._signed_files = {
            (KeyPurpose.UEFI, "hello/boot/vmlinuz"): signed_vmlinuz_path
        }

        self.task._add_signed_files(code_signing_path, source_path)

        self.assertEqual(
            (
                source_path
                / "debian"
                / "signatures"
                / "hello"
                / "boot"
                / "vmlinuz.sig"
            ).read_text(),
            "signed vmlinuz\n",
        )

    def test_build_source_package_dpkg_source_error(self) -> None:
        """_build_source_package raises if dpkg-source exits non-zero."""
        self.patch_executor_methods()
        self.task._remote_execute_directory = self.create_temporary_directory()
        local_execute_directory = self.create_temporary_directory()
        local_source_path = local_execute_directory / "source"
        local_source_path.mkdir()

        with (
            self.assertRaisesRegex(
                AssembleError, "dpkg-source exited with code 1"
            ),
            mock.patch.object(self.task, "run_cmd", return_value=1),
        ):
            self.task._build_source_package(local_source_path)

    def test_build_source_package_dpkg_genchanges_error(self) -> None:
        """_build_source_package raises if dpkg-genchanges exits non-zero."""
        self.patch_executor_methods()
        self.task._remote_execute_directory = self.create_temporary_directory()
        local_execute_directory = self.create_temporary_directory()
        local_source_path = local_execute_directory / "source"
        self._create_source_template(local_source_path)
        now = dt.datetime.now(dt.UTC).strftime("%a, %d %b %Y %H:%M:%S %z")
        (local_source_path / "debian/changelog").write_text(
            dedent(
                f"""\
                hello (1.0-1) unstable; urgency=low

                  * Initial release.

                 -- Example Maintainer <example@example.org>  {now}
                """
            )
        )

        with (
            self.assertRaisesRegex(
                AssembleError, "dpkg-genchanges exited with code 1"
            ),
            mock.patch.object(self.task, "run_cmd", side_effect=[0, 1]),
        ):
            self.task._build_source_package(local_source_path)

    def test_build_source_package_success(self) -> None:
        """_build_source_package runs the correct commands."""
        self.patch_executor_methods()
        self.task._remote_execute_directory = self.create_temporary_directory()
        local_execute_directory = self.create_temporary_directory()
        remote_source_path = self.task._remote_execute_directory / "source"
        local_source_path = local_execute_directory / "source"
        self._create_source_template(local_source_path)
        now = dt.datetime.now(dt.UTC).strftime("%a, %d %b %Y %H:%M:%S %z")
        (local_source_path / "debian/changelog").write_text(
            dedent(
                f"""\
                hello (1.0-1) unstable; urgency=low

                  * Initial release.

                 -- Example Maintainer <example@example.org>  {now}
                """
            )
        )

        with (
            self.assertLogs("debusine.tasks", level=logging.INFO) as log,
            mock.patch.object(
                self.task, "run_cmd", return_value=0
            ) as mock_run_cmd,
        ):
            self.task._build_source_package(local_source_path)

        expected_changes_path = (
            remote_source_path.parent / "hello_1.0-1_source.changes"
        )
        mock_run_cmd.assert_has_calls(
            [
                mock.call(["dpkg-source", "-b", "."], remote_source_path),
                mock.call(
                    [
                        "dpkg-genchanges",
                        "-S",
                        "-DDistribution=unstable",
                        "-UCloses",
                        f"-O{expected_changes_path}",
                    ],
                    remote_source_path,
                ),
            ]
        )
        self.assertEqual(
            log.output,
            [
                "INFO:debusine.tasks:Executing: dpkg-source -b .",
                "INFO:debusine.tasks:dpkg-source exited with code 0",
                f"INFO:debusine.tasks:Executing: "
                f"dpkg-genchanges -S -DDistribution=unstable -UCloses "
                f"-O{expected_changes_path}",
                "INFO:debusine.tasks:dpkg-genchanges exited with code 0",
            ],
        )

    def test_create_artifacts(self) -> None:
        """_create_artifacts creates artifacts for a source package."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        self.assertTrue(self.task.configure_for_execution(download_directory))
        assert isinstance(self.task.executor_instance, mock.MagicMock)
        self.task.executor_instance.file_pull.side_effect = shutil.copy2

        remote_execute_directory = self.create_temporary_directory()
        self.task._local_execute_directory = self.create_temporary_directory()
        remote_dsc_path = remote_execute_directory / "hello-signed_1.0+1.dsc"
        remote_changes_path = (
            remote_execute_directory / "hello-signed_1.0+1_source.changes"
        )
        ArtifactPlayground.write_deb822_file(
            deb822.Dsc,
            remote_dsc_path,
            [],
            source="hello-signed",
            version="1.0+1",
        )
        ArtifactPlayground.write_deb822_file(
            deb822.Changes,
            remote_changes_path,
            [remote_dsc_path],
            source="hello-signed",
            version="1.0+1",
        )

        self.task._create_artifacts(remote_changes_path)

        self.assertEqual(
            self.task._source_package_artifact,
            SourcePackage.create(
                name="hello-signed",
                version="1.0+1",
                files=[
                    self.task._local_execute_directory / remote_dsc_path.name
                ],
            ),
        )
        self.assertEqual(
            self.task._upload_artifact,
            Upload.create(
                changes_file=(
                    self.task._local_execute_directory
                    / remote_changes_path.name
                )
            ),
        )

    def test_upload_artifacts_execution_failed(self) -> None:
        """upload_artifacts does not upload anything if execution failed."""
        debusine_mock = self.mock_debusine()
        execute_directory = self.create_temporary_directory()

        self.task.upload_artifacts(
            execute_directory, execution_result=WorkRequestResults.FAILURE
        )

        debusine_mock.upload_artifact.assert_not_called()
        debusine_mock.relation_create.assert_not_called()

    def test_upload_artifacts_execution_succeeded(self) -> None:
        """upload_artifacts uploads if execution succeeded."""
        directory = self.create_temporary_directory()
        dsc_path = directory / "hello-signed_1.0+1.dsc"
        changes_path = directory / "hello-signed_1.0+1_source.changes"
        ArtifactPlayground.write_deb822_file(
            deb822.Dsc, dsc_path, [], source="hello-signed", version="1.0+1"
        )
        ArtifactPlayground.write_deb822_file(
            deb822.Changes,
            changes_path,
            [dsc_path],
            source="hello-signed",
            version="1.0+1",
        )
        self.task._built_using_ids = [3]
        self.task._source_package_artifact = SourcePackage.create(
            name="hello-signed",
            version="1.0+1",
            files=[dsc_path],
        )
        self.task._upload_artifact = Upload.create(changes_file=changes_path)
        debusine_mock = self.mock_debusine()
        debusine_mock.upload_artifact.side_effect = [
            create_remote_artifact(id=20, workspace=self.task.workspace_name),
            create_remote_artifact(id=21, workspace=self.task.workspace_name),
        ]
        execute_directory = self.create_temporary_directory()

        self.task.upload_artifacts(
            execute_directory, execution_result=WorkRequestResults.SUCCESS
        )

        debusine_mock.upload_artifact.assert_has_calls(
            [
                mock.call(
                    self.task._source_package_artifact,
                    workspace=self.task.workspace_name,
                    work_request=self.task.work_request_id,
                ),
                mock.call(
                    self.task._upload_artifact,
                    workspace=self.task.workspace_name,
                    work_request=self.task.work_request_id,
                ),
            ]
        )
        debusine_mock.relation_create.assert_has_calls(
            [mock.call(20, 3, "built-using"), mock.call(21, 20, "extends")]
        )

    def test_cleanup_without_local_execute_directory(self) -> None:
        """cleanup() works even if _local_execute_directory hasn't been set."""
        # Just check that it can be called without raising an exception.
        self.task.cleanup()

    def test_execute_end_to_end(self) -> None:
        """End-to-end execution works."""
        self.task.dynamic_data = AssembleSignedSourceDynamicData(
            environment_id=1, template_id=11, signed_ids=[12]
        )

        # Create some test files.
        directory = self.create_temporary_directory()
        template_deb_path = directory / "hello-signed_1.0+1_amd64.deb"
        code_signing_path = PurePath("usr/share/code-signing/hello-signed")
        source_template_path = code_signing_path / "source-template"
        now = dt.datetime.now(dt.UTC).strftime("%a, %d %b %Y %H:%M:%S %z")
        ArtifactPlayground.write_deb_file(
            template_deb_path,
            source_name="hello-signed",
            source_version="1.0+1",
            data_files={
                (code_signing_path / "files.json"): json.dumps(
                    {
                        "packages": {
                            "hello": {
                                "files": [
                                    {"sig_type": "efi", "file": "boot/vmlinuz"}
                                ]
                            }
                        }
                    }
                ).encode(),
                (
                    source_template_path / "debian/source/format"
                ): b"3.0 (native)\n",
                source_template_path / "debian/changelog": dedent(
                    f"""\
                    hello-signed (1.0+1) unstable; urgency=low

                      * Initial release.

                     -- Example Maintainer <example@example.org>  {now}
                    """
                ).encode(),
                source_template_path / "debian/control": dedent(
                    """\
                    Source: hello-signed
                    Section: admin
                    Priority: optional
                    Maintainer: Example Maintainer <example@example.org>

                    Package: hello-signed
                    Architecture: amd64
                    Description: Example package
                    """
                ).encode(),
                source_template_path / "debian/copyright": b"",
                source_template_path / "debian/rules": b"",
            },
        )
        signed_vmlinuz_path = directory / "vmlinuz.sig"
        signed_vmlinuz_path.write_text("signed vmlinuz\n")

        # Set up a mock client.
        debusine_mock = self.mock_debusine()
        artifacts = {
            11: self._create_binary_package_response(11),
            12: create_artifact_response(
                id=12,
                category=ArtifactCategory.SIGNING_OUTPUT,
                data={
                    "purpose": "uefi",
                    "fingerprint": "0" * 64,
                    "results": [
                        {
                            "file": "hello/boot/vmlinuz",
                            "output_file": "hello/boot/vmlinuz.sig",
                        }
                    ],
                },
                files={"hello/boot/vmlinuz.sig": create_file_response()},
            ),
        }
        artifact_files = {
            (11, "hello-signed_1.0+1_amd64.deb"): template_deb_path,
            (12, "hello/boot/vmlinuz.sig"): signed_vmlinuz_path,
        }
        artifact_categories = {
            3: ArtifactCategory.BINARY_PACKAGE,
            4: ArtifactCategory.SIGNING_INPUT,
        }
        relations = {4: [(3, "relates-to")], 12: [(4, "relates-to")]}
        uploaded_artifact_ids = {
            ArtifactCategory.SOURCE_PACKAGE: 20,
            ArtifactCategory.UPLOAD: 21,
            ArtifactCategory.WORK_REQUEST_DEBUG_LOGS: 22,
        }
        signed_tarball_path = directory / "hello-signed_1.0+1.tar.xz"

        def download_artifact(
            artifact_id: int,
            destination: Path,
            tarball: bool = False,
            workspace: str | None = None,  # noqa: ARG001, U100
        ) -> ArtifactResponse:
            assert not tarball
            for file_name in artifacts[artifact_id].files:
                target = destination / file_name
                target.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy2(artifact_files[(artifact_id, file_name)], target)
            return artifacts[artifact_id]

        def relation_list(artifact_id: int) -> RelationsResponse:
            return RelationsResponse.model_validate(
                [
                    {
                        "id": 1,
                        "artifact": artifact_id,
                        "target": target_id,
                        "type": relation_type,
                    }
                    for target_id, relation_type in relations[artifact_id]
                ]
            )

        def artifact_get(
            artifact_id: int,
            workspace: str,  # noqa: ARG001, U100
        ) -> ArtifactResponse:
            return create_artifact_response(
                id=6, category=artifact_categories[artifact_id]
            )

        def upload_artifact(
            local_artifact: LocalArtifact[Any], **kwargs: Any
        ) -> RemoteArtifact:
            if local_artifact.category == ArtifactCategory.SOURCE_PACKAGE:
                self.assertCountEqual(
                    local_artifact.files.keys(),
                    ["hello-signed_1.0+1.dsc", "hello-signed_1.0+1.tar.xz"],
                )
                shutil.copy(
                    local_artifact.files["hello-signed_1.0+1.tar.xz"],
                    signed_tarball_path,
                )
            return create_remote_artifact(
                id=uploaded_artifact_ids[local_artifact.category],
                workspace="testing",
            )

        debusine_mock.download_artifact.side_effect = download_artifact
        debusine_mock.relation_list.side_effect = relation_list
        debusine_mock.artifact_get.side_effect = artifact_get
        debusine_mock.upload_artifact.side_effect = upload_artifact

        def run(
            args: list[str],
            run_as_root: bool = False,  # noqa: ARG001, U100
            **kwargs: Any,
        ) -> subprocess.CompletedProcess[AnyStr]:
            if args[0] == "apt-get":
                return subprocess.CompletedProcess(args, 0)
            else:
                kwargs.setdefault("stdout", subprocess.PIPE)
                kwargs.setdefault("stderr", subprocess.PIPE)
                return subprocess.run(args, **kwargs)

        with (
            mock.patch.object(
                self.task, "_prepare_executor_instance", autospec=True
            ),
            mock.patch.object(
                self.task, "executor_instance", autospec=True
            ) as mock_executor_instance,
        ):
            mock_executor_instance.file_pull.side_effect = shutil.copy2
            mock_executor_instance.file_push.side_effect = shutil.copy2
            mock_executor_instance.run.side_effect = run
            self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        self.assertEqual(debusine_mock.upload_artifact.call_count, 3)
        with tarfile.open(signed_tarball_path) as tar:
            vmlinuz_sig_file = tar.extractfile(
                "source/debian/signatures/hello/boot/vmlinuz.sig"
            )
            assert vmlinuz_sig_file is not None
            self.assertEqual(vmlinuz_sig_file.read(), b"signed vmlinuz\n")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the Executor basic interface."""

import os
from pathlib import Path, PurePath
from subprocess import CalledProcessError, CompletedProcess, PIPE
from typing import Any, AnyStr, Literal, override
from unittest.mock import Mock, call, patch

import debusine.worker.tags as wtags
from debusine.artifacts.models import RuntimeStatistics
from debusine.tasks.executors.base import (
    ExecutorInterface,
    ExecutorStatistics,
    InstanceInterface,
    InstanceNotRunning,
    InstanceRunning,
    _backends,
    analyze_worker_all_executors,
    executor_backends,
    executor_class,
    executor_provided_tags,
)
from debusine.test import TestCase


class ExecutorStatisticsTests(TestCase):
    """Tests for :py:class:`ExecutorStatistics`."""

    def test_init(self) -> None:
        s = ExecutorStatistics()
        self.assertIsNone(s.disk_space)
        self.assertIsNone(s.memory)
        self.assertIsNone(s.available_disk_space)
        self.assertIsNone(s.available_memory)
        self.assertIsNone(s.cpu_count)

    def test_add_to(self) -> None:
        ES = ExecutorStatistics
        RS = RuntimeStatistics
        for es, rs, expected in (
            (ES(), RS(), RS()),
            (
                ES(
                    disk_space=1,
                    memory=2,
                    available_disk_space=4,
                    available_memory=8,
                    cpu_count=16,
                ),
                RS(),
                RS(
                    disk_space=1,
                    memory=2,
                    available_disk_space=4,
                    available_memory=8,
                    cpu_count=16,
                ),
            ),
            (ES(disk_space=1), RS(disk_space=4), RS(disk_space=4)),
            (ES(disk_space=4), RS(disk_space=1), RS(disk_space=4)),
            (ES(disk_space=None), RS(disk_space=1), RS(disk_space=1)),
            (ES(memory=1), RS(memory=4), RS(memory=4)),
            (ES(memory=4), RS(memory=1), RS(memory=4)),
            (ES(memory=None), RS(memory=1), RS(memory=1)),
            (
                ES(available_disk_space=1),
                RS(available_disk_space=4),
                RS(available_disk_space=4),
            ),
            (
                ES(available_disk_space=4),
                RS(available_disk_space=1),
                RS(available_disk_space=1),
            ),
            (
                ES(available_disk_space=None),
                RS(available_disk_space=1),
                RS(available_disk_space=1),
            ),
            (
                ES(available_memory=1),
                RS(available_memory=4),
                RS(available_memory=4),
            ),
            (
                ES(available_memory=4),
                RS(available_memory=1),
                RS(available_memory=1),
            ),
            (
                ES(available_memory=None),
                RS(available_memory=1),
                RS(available_memory=1),
            ),
            (ES(cpu_count=1), RS(cpu_count=4), RS(cpu_count=4)),
            (ES(cpu_count=4), RS(cpu_count=1), RS(cpu_count=1)),
            (ES(cpu_count=None), RS(cpu_count=1), RS(cpu_count=1)),
        ):
            with self.subTest(es=str(es), rs=str(rs)):
                actual = rs.model_copy()
                es.add_to(actual)
                self.assertEqual(actual, expected)


class SampleInstance(InstanceInterface):
    """Instance used for testing."""

    started: bool = False

    @override
    def is_started(self) -> bool:
        """Return the started status."""
        return self.started

    @override
    def do_start(self) -> None:
        """Not implemented."""
        raise NotImplementedError()

    @override
    def do_stop(self) -> None:
        """Not implemented."""
        raise NotImplementedError()

    @override
    def do_restart(self) -> None:
        """Not implemented."""
        raise NotImplementedError()

    @override
    def do_file_push(
        self, source: Path, target: PurePath, uid: int, gid: int, mode: int
    ) -> None:
        """Not implemented."""
        raise NotImplementedError()

    @override
    def do_file_pull(self, source: PurePath, target: Path) -> None:
        """Not implemented."""
        raise NotImplementedError()

    @override
    def do_run(
        self,
        args: list[str],
        text: Literal[True] | Literal[False] | None = None,
        cwd: str | os.PathLike[str] | None = None,
        run_as_root: bool = False,
        **kwargs: Any,
    ) -> CompletedProcess[AnyStr]:
        """Not implemented."""
        raise NotImplementedError()


class ExecutorTests(TestCase):
    """Unit tests for the Executor basic interface."""

    def test_backends_populated(self) -> None:
        """Test that we have populated some backends."""
        self.assertGreater(len(_backends), 0)

    def test_analyze_worker_all_executors_no_backends_available(self) -> None:
        """analyze_worker_all_executors copes with no available backends."""
        self.mock_is_command_available({})
        executors_metadata = analyze_worker_all_executors()
        self.assertNotEqual(executors_metadata, {})
        for key, value in executors_metadata.items():
            self.assertRegex(key, r"^executor:.*:available$")
            self.assertFalse(value)

    def test_analyze_worker_all_executors_unshare_available(self) -> None:
        """analyze_worker_all_executors reports available backends."""
        self.mock_is_command_available({"unshare": True, "ip": True})
        executors_metadata = analyze_worker_all_executors()
        self.assertTrue(executors_metadata["executor:unshare:available"])

    def test_executor_class_and_backends(self) -> None:
        backends = executor_backends()
        self.assertGreater(len(backends), 0)
        instance = executor_class(backends[0])
        self.assertTrue(issubclass(instance, ExecutorInterface))

    def test_executor_provided_tags_no_backends_available(self) -> None:
        self.mock_is_command_available({})
        self.assertCountEqual(executor_provided_tags(), [])

    def test_executor_provided_tags_unshare_available(self) -> None:
        self.mock_is_command_available({"unshare": True, "ip": True})
        self.assertCountEqual(
            executor_provided_tags(), [wtags.EXECUTOR_PREFIX + "unshare"]
        )


class InstanceTests(TestCase):
    """Unit tests for the Instance basic interface."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.instance = SampleInstance()

    def patch_instance(self, method: str) -> Mock:
        """Mock method on self.instance."""
        patcher = patch.object(self.instance, method, autospec=True)
        self.addCleanup(patcher.stop)
        return patcher.start()

    def test_start(self) -> None:
        """Test that start calls do_start()."""
        do_start = self.patch_instance("do_start")
        self.instance.start()
        do_start.assert_called_once_with()

    def test_start_started(self) -> None:
        """Test that start requires a stopped instance."""
        self.instance.started = True
        with self.assertRaisesRegex(
            InstanceRunning, "Cannot start a started instance"
        ):
            self.instance.start()

    def test_stop(self) -> None:
        """Test that stop calls do_stop()."""
        self.instance.started = True
        do_stop = self.patch_instance("do_stop")
        self.instance.stop()
        do_stop.assert_called_once_with()

    def test_stop_stopped(self) -> None:
        """Test that stop requires a started instance."""
        with self.assertRaisesRegex(
            InstanceNotRunning, "Cannot stop a stopped instance"
        ):
            self.instance.stop()

    def test_restart(self) -> None:
        """Test that restart calls do_restart()."""
        self.instance.started = True
        do_restart = self.patch_instance("do_restart")
        self.instance.restart()
        do_restart.assert_called_once_with()

    def test_restart_stopped(self) -> None:
        """Test that restart requires a started instance."""
        with self.assertRaisesRegex(
            InstanceNotRunning, "Cannot restart a stopped instance"
        ):
            self.instance.restart()

    def test_file_push(self) -> None:
        """Test that file_push calls do_file_push()."""
        self.instance.started = True
        do_file_push = self.patch_instance("do_file_push")
        source = Path("source")
        dest = Path("/dest")
        self.instance.file_push(source, dest)
        do_file_push.assert_called_once_with(source, dest, 0, 0, 0o644)

    def test_file_push_stopped(self) -> None:
        """Test that file_push requires a started instance."""
        with self.assertRaisesRegex(
            InstanceNotRunning, "Cannot push files to a stopped instance"
        ):
            self.instance.file_push(Path("source"), Path("/dest"))

    def test_file_push_relative(self) -> None:
        """Test that file_push requires a relative destination."""
        self.instance.started = True
        with self.assertRaisesRegex(
            ValueError, "target must be an absolute path"
        ):
            self.instance.file_push(Path("source"), Path("dest"))

    def test_file_pull(self) -> None:
        """Test that file_pull calls do_file_pull()."""
        self.instance.started = True
        do_file_pull = self.patch_instance("do_file_pull")
        source = Path("/source")
        dest = Path("dest")
        self.instance.file_pull(source, dest)
        do_file_pull.assert_called_once_with(source, dest)

    def test_file_pull_stopped(self) -> None:
        """Test that file_pull requires a started instance."""
        with self.assertRaisesRegex(
            InstanceNotRunning, "Cannot pull files from a stopped instance"
        ):
            self.instance.file_pull(Path("/source"), Path("dest"))

    def test_file_pull_relative(self) -> None:
        """Test that file_pull requires a relative source."""
        self.instance.started = True
        with self.assertRaisesRegex(
            ValueError, "source must be an absolute path"
        ):
            self.instance.file_pull(Path("source"), Path("dest"))

    def test_directory_push(self) -> None:
        """Test that directory_push calls do_directory_push()."""
        self.instance.started = True
        do_directory_push = self.patch_instance("do_directory_push")
        source = Path("source")
        dest = Path("/dest")
        self.instance.directory_push(source, dest)
        do_directory_push.assert_called_once_with(source, dest, 0, 0)

    def test_directory_push_stopped(self) -> None:
        """Test that directory_push requires a started instance."""
        with self.assertRaisesRegex(
            InstanceNotRunning, "Cannot push files to a stopped instance"
        ):
            self.instance.directory_push(Path("source"), Path("/dest"))

    def test_directory_push_relative(self) -> None:
        """Test that directory_push requires a relative destination."""
        self.instance.started = True
        with self.assertRaisesRegex(
            ValueError, "target must be an absolute path"
        ):
            self.instance.directory_push(Path("source"), Path("dest"))

    def test_do_directory_push(self) -> None:
        """do_directory_push creates the directories and copies the files."""
        self.instance.started = True

        temp_directory = self.create_temporary_directory()
        (source_directory := temp_directory / "source").mkdir()
        source_directory.chmod(0o755)
        (sub_directory := source_directory / "sub-directory").mkdir()
        sub_directory.chmod(0o755)
        (sub_sub_directory := sub_directory / "sub-sub-directory").mkdir()
        sub_sub_directory.chmod(0o755)

        (file1 := source_directory / "file1.txt").touch()
        file1.chmod(0o640)
        (file2 := sub_directory / "file2.sh").touch()
        file2.chmod(0o755)

        target = PurePath("/somewhere")

        mkdir_mock = self.patch_instance("do_mkdir")
        do_file_push_mock = self.patch_instance("do_file_push")
        self.instance.do_directory_push(source_directory, target, 101, 102)

        mkdir_mock.assert_has_calls(
            [
                call(PurePath("/somewhere/source"), 101, 102, 0o755, False),
                call(
                    PurePath("/somewhere/source/sub-directory"),
                    101,
                    102,
                    0o755,
                    False,
                ),
                call(
                    PurePath(
                        "/somewhere/source/sub-directory/sub-sub-directory"
                    ),
                    101,
                    102,
                    0o755,
                    False,
                ),
            ]
        )
        do_file_push_mock.assert_has_calls(
            [
                call(
                    Path(source_directory / "file1.txt"),
                    PurePath("/somewhere/source/file1.txt"),
                    101,
                    102,
                    0o640,
                ),
                call(
                    Path(source_directory / "sub-directory/file2.sh"),
                    PurePath("/somewhere/source/sub-directory/file2.sh"),
                    101,
                    102,
                    0o755,
                ),
            ]
        )

    def test_run(self) -> None:
        """Test that run calls do_run()."""
        self.instance.started = True
        do_run = self.patch_instance("do_run")
        self.instance.run(["cmd", "--arg"])
        do_run.assert_called_once_with(
            ["cmd", "--arg"], stderr=PIPE, stdout=PIPE
        )

    def test_run_stopped(self) -> None:
        """Test that run requires a started instance."""
        with self.assertRaisesRegex(
            InstanceNotRunning, "Cannot run commands on a stopped instance"
        ):
            self.instance.run(["foo"])

    def test_run_check_output(self) -> None:
        """Test that _run_check_output calls run()."""
        with patch.object(self.instance, "run") as run_mock:
            self.instance._run_check_output(["true"])
        run_mock.assert_called_once_with(
            ["true"],
            check=True,
            run_as_root=True,
            stderr=PIPE,
            stdout=PIPE,
            text=True,
        )

    def test_get_uid_uncached(self) -> None:
        """Test _get_uid when not cached."""
        with patch.object(self.instance, "_run_check_output") as run_mock:
            run_mock.return_value = "testuser:x:101:105:Test User:/foo:/bin/sh"
            self.assertEqual(self.instance._get_uid("testuser"), 101)
        run_mock.assert_called_once_with(["getent", "passwd", "testuser"])
        self.assertEqual(self.instance._uid_map["testuser"], 101)

    def test_get_uid_cached(self) -> None:
        """Test _get_uid when cached."""
        with patch.object(self.instance, "_run_check_output") as run_mock:
            self.assertEqual(self.instance._get_uid("root"), 0)
        run_mock.assert_not_called()

    def test_get_uid_unknown(self) -> None:
        """Test _get_uid when not existent."""
        with patch.object(self.instance, "_run_check_output") as run_mock:
            run_mock.side_effect = CalledProcessError(
                cmd=["getent", "passwd", "testuser"], returncode=2
            )
            with self.assertRaisesRegex(
                KeyError, "User not found in instance: testuser"
            ):
                self.instance._get_uid("testuser")
        self.assertNotIn("testuser", self.instance._uid_map)

    def test_get_uid_unexpected_error(self) -> None:
        """Test _get_uid when not existent."""
        with patch.object(self.instance, "_run_check_output") as run_mock:
            run_mock.side_effect = CalledProcessError(
                cmd=["getent", "passwd", "testuser"], returncode=1
            )
            with self.assertRaises(CalledProcessError):
                self.instance._get_uid("testuser")

    def test_get_gid_uncached(self) -> None:
        """Test _get_gid when not cached."""
        with patch.object(self.instance, "_run_check_output") as run_mock:
            run_mock.return_value = "testgroup:x:102:"
            self.assertEqual(self.instance._get_gid("testgroup"), 102)
        run_mock.assert_called_once_with(["getent", "group", "testgroup"])
        self.assertEqual(self.instance._gid_map["testgroup"], 102)

    def test_get_gid_cached(self) -> None:
        """Test _get_gid when cached."""
        with patch.object(self.instance, "_run_check_output") as run_mock:
            self.assertEqual(self.instance._get_gid("root"), 0)
        run_mock.assert_not_called()

    def test_get_gid_unknown(self) -> None:
        """Test _get_gid when not existent."""
        with patch.object(self.instance, "_run_check_output") as run_mock:
            run_mock.side_effect = CalledProcessError(
                cmd=["getent", "group", "testgroup"], returncode=2
            )
            with self.assertRaisesRegex(
                KeyError, "Group not found in instance: testgroup"
            ):
                self.instance._get_gid("testgroup")
        self.assertNotIn("testgroup", self.instance._gid_map)

    def test_get_gid_unexpected_error(self) -> None:
        """Test _get_gid when not existent."""
        with patch.object(self.instance, "_run_check_output") as run_mock:
            run_mock.side_effect = CalledProcessError(
                cmd=["getent", "group", "testgroup"], returncode=1
            )
            with self.assertRaises(CalledProcessError):
                self.instance._get_gid("testgroup")

    def test_create_user_exists(self) -> None:
        """Test create_user when user exists."""
        self.instance._uid_map["testuser"] = 42
        with patch.object(self.instance, "_run_check_output") as run_mock:
            self.assertEqual(self.instance.create_user("testuser"), 42)
        run_mock.assert_not_called()

    def test_create_user_not_exists(self) -> None:
        """Test create_user when user doesn't exist."""

        def create_user(args: list[str]) -> str:
            """Fake for run_check_output that creates a user."""
            if args[0] == "sh":
                return ""
            if args[0] == "useradd":
                self.instance._uid_map["testuser"] = 42
                return ""
            raise CalledProcessError(cmd=args, returncode=2)

        with patch.object(self.instance, "_run_check_output") as run_mock:
            run_mock.side_effect = create_user
            self.assertEqual(self.instance.create_user("testuser"), 42)
        run_mock.assert_has_calls(
            [
                call(["getent", "passwd", "testuser"]),
                call(
                    [
                        "sh",
                        "-c",
                        (
                            "if [ ! -x /usr/sbin/useradd ]; then "
                            "apt-get update && apt-get -y install passwd; "
                            "fi"
                        ),
                    ]
                ),
                call(["useradd", "testuser"]),
            ]
        )

    def test_create_user_default_username(self) -> None:
        """Test create_user when no username is supplied."""
        self.instance._uid_map["_debusine"] = 42
        with patch.object(self.instance, "_run_check_output") as run_mock:
            self.assertEqual(self.instance.create_user(), 42)
        run_mock.assert_not_called()

    def test_mkdir(self) -> None:
        """Test that mkdir calls do_mkdir()."""
        self.instance.started = True
        do_mkdir = self.patch_instance("do_mkdir")
        dest = Path("/dest")
        self.instance.mkdir(dest)
        do_mkdir.assert_called_once_with(dest, 0, 0, 0o755, False)

    def test_mkdir_stopped(self) -> None:
        """Test that mkdir requires a started instance."""
        with self.assertRaisesRegex(
            InstanceNotRunning, "Cannot make directories in a stopped instance"
        ):
            self.instance.mkdir(Path("/dest"))

    def test_mkdir_relative(self) -> None:
        """Test that mkdir requires a relative target."""
        self.instance.started = True
        with self.assertRaisesRegex(
            ValueError, "target must be an absolute path"
        ):
            self.instance.mkdir(Path("dest"))

    def test_do_mkdir(self) -> None:
        """do_mkdir() use run() as expected."""
        uid = 101
        gid = 102
        target = PurePath("/etc/dir1")

        with patch.object(self.instance, "_run_check_output") as run_mocked:
            self.instance.do_mkdir(target, uid, gid, 0o755, False)

        run_expected_calls = [
            call(["mkdir", "--mode=755", str(target)]),
            call(["chown", f"{uid}:{gid}", "/etc/dir1"]),
        ]
        run_mocked.assert_has_calls(run_expected_calls)

    def test_do_mkdir_parents(self) -> None:
        """do_mkdir() passes --parents."""
        uid = 101
        gid = 102
        target = PurePath("/etc/dir1")

        with patch.object(self.instance, "_run_check_output") as run_mocked:
            self.instance.do_mkdir(target, uid, gid, 0o755, parents=True)

        run_expected_calls = [
            call(["mkdir", "--mode=755", "--parents", str(target)]),
            call(["chown", f"{uid}:{gid}", "/etc/dir1"]),
        ]
        run_mocked.assert_has_calls(run_expected_calls)

    def test_do_mkdir_root(self) -> None:
        """do_mkdir() skips chown for root."""
        target = PurePath("/etc/dir1")

        with patch.object(self.instance, "_run_check_output") as run_mocked:
            self.instance.do_mkdir(target, 0, 0, 0o755, False)

        run_mocked.assert_called_once_with(["mkdir", "--mode=755", str(target)])

    def test_do_mkdir_exists(self) -> None:
        """do_mkdir() raise FileExistsError."""
        self.instance.started = True
        target = PurePath("/some/directory")

        with patch.object(self.instance, "_run_check_output") as run_mock:
            run_mock.side_effect = CalledProcessError(
                cmd=["mkdir"],
                returncode=1,
                stderr=(
                    f"mkdir: cannot create directory ‘{target}’: File exists"
                ),
            )
            msg = rf'^Cannot create directory: "{target}" already exists$'
            with self.assertRaisesRegex(FileExistsError, msg):
                self.instance.do_mkdir(target, 0, 0, 0o755, False)

    def test_do_mkdir_unexpected_error(self) -> None:
        """do_mkdir() raise CalledProcessError on an unknown error."""
        self.instance.started = True
        target = PurePath("/some/directory")

        with patch.object(self.instance, "_run_check_output") as run_mock:
            run_mock.side_effect = CalledProcessError(
                cmd=["mkdir"], returncode=1, stderr="mkdir: Something Happened!"
            )
            with self.assertRaises(CalledProcessError):
                self.instance.do_mkdir(target, 0, 0, 0o755, False)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the asset views."""

import base64
from typing import Any, override

from django.urls import reverse
from django.utils.dateparse import parse_datetime
from rest_framework import status

from debusine.artifacts.models import CollectionCategory
from debusine.assets import AssetCategory, KeyPurpose, SigningKeyData
from debusine.db.models import (
    Artifact,
    Asset,
    AssetUsage,
    Collection,
    Scope,
    Token,
    WorkRequest,
    Worker,
    Workspace,
)
from debusine.db.models.assets import AssetUsageRoles
from debusine.db.models.scopes import ScopeRoles
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.db.playground.playground import SourceUploadArtifacts
from debusine.server.views.assets import AssetPermissionCheckView, log
from debusine.signing.models import SigningMode
from debusine.signing.tasks.models import (
    DebsignData,
    DebsignDynamicData,
    SignData,
    SignDynamicData,
    SignRepositoryIndexData,
    SignRepositoryIndexDynamicData,
)
from debusine.tasks.models import LookupMultiple, WorkerType
from debusine.test.django import TestCase, TestResponseType


class AssetViewTests(TestCase):
    """Tests for AssetView."""

    scenario = scenarios.DefaultContextAPI()
    playground_memory_file_store = False

    def post_asset_create(
        self,
        *,
        token_key: str,
        data: dict[str, Any],
        scope: Scope,
    ) -> TestResponseType:
        """
        Call self.client.post() to api:asset.

        Pass the token.key and content_type="application/json".

        :param token_key: value of the Token header for the request
        :param data: data to post
        """
        headers: dict[str, str] = {}
        headers["X-Debusine-Scope"] = scope.name

        return self.client.post(
            reverse("api:asset"),
            data=data,
            HTTP_TOKEN=token_key,
            content_type="application/json",
            headers=headers,
        )

    def get_asset_list(
        self,
        *,
        token_key: str,
        query_params: dict[str, Any],
        scope: Scope,
    ) -> TestResponseType:
        """
        Call self.client.get() to api:asset.

        Pass the token.key and content_type="application/json".

        :param token_key: value of the Token header for the request
        :param query_params: query keys and values
        """
        headers: dict[str, str] = {}
        headers["Accept"] = "application/json"
        headers["X-Debusine-Scope"] = scope.name

        return self.client.get(
            reverse("api:asset"),
            HTTP_TOKEN=token_key,
            headers=headers,
            data=query_params,
        )

    def signing_key_asset_as_dict_request(
        self,
        fingerprint: str,
        public_key: str,
        workspace: Workspace,
        description: str | None = None,
        purpose: KeyPurpose = KeyPurpose.OPENPGP_UPLOAD,
        work_request_id: int | None = None,
    ) -> dict[str, Any]:
        """Generate an asset creation request dict."""
        return {
            "workspace": workspace.name,
            "category": AssetCategory.SIGNING_KEY,
            "work_request": work_request_id,
            "data": SigningKeyData(
                purpose=purpose,
                fingerprint=fingerprint,
                public_key=base64.b64encode(public_key.encode()).decode(),
                description=description,
            ).model_dump(),
        }

    def test_create_asset(self) -> None:
        """Create a signing-key asset."""
        workspace = self.playground.get_default_workspace()
        user = self.playground.get_default_user()
        self.playground.create_group_role(
            workspace.scope, ScopeRoles.OWNER, users=[user]
        )
        response = self.post_asset_create(
            token_key=self.scenario.user_token.key,
            data=self.signing_key_asset_as_dict_request(
                fingerprint="ABC123",
                public_key="PUBLIC KEY",
                workspace=workspace,
            ),
            scope=workspace.scope,
        )
        body = self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        asset = Asset.objects.get(id=body["id"])
        self.assertEqual(asset.category, AssetCategory.SIGNING_KEY)
        data = asset.data_model
        assert isinstance(data, SigningKeyData)
        self.assertEqual(data.fingerprint, "ABC123")
        self.assertEqual(data.public_key, b"PUBLIC KEY")
        self.assertEqual(asset.workspace, workspace)
        self.assertEqual(asset.created_by, user)
        self.assertIsNone(asset.created_by_work_request)
        self.assertTrue(asset.usage.filter(workspace=workspace).exists())

    def test_worker_create_asset(self) -> None:
        """Worker creates a signing-key asset."""
        workspace = self.playground.get_default_workspace()
        user = self.playground.get_default_user()
        self.playground.create_group_role(
            workspace.scope, ScopeRoles.OWNER, users=[user]
        )
        worker = self.playground.create_worker(worker_type=WorkerType.SIGNING)
        work_request = self.playground.create_worker_task(
            worker=worker,
            mark_running=True,
            workspace=workspace,
            created_by=user,
        )
        token = self.playground.create_work_request_token(work_request)
        response = self.post_asset_create(
            token_key=token.key,
            data=self.signing_key_asset_as_dict_request(
                fingerprint="ABC123",
                public_key="PUBLIC KEY",
                workspace=workspace,
                work_request_id=work_request.id,
            ),
            scope=workspace.scope,
        )
        body = self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        asset = Asset.objects.get(id=body["id"])
        self.assertEqual(asset.category, AssetCategory.SIGNING_KEY)
        data = asset.data_model
        assert isinstance(data, SigningKeyData)
        self.assertEqual(data.fingerprint, "ABC123")
        self.assertEqual(data.public_key, b"PUBLIC KEY")
        self.assertEqual(asset.workspace, workspace)
        self.assertIsNone(asset.created_by)
        self.assertEqual(asset.created_by_work_request, work_request)
        self.assertTrue(asset.usage.filter(workspace=workspace).exists())

    def test_create_asset_idempotent(self) -> None:
        """Create a signing-key asset that already exists, return 200."""
        workspace = self.playground.get_default_workspace()
        user = self.playground.get_default_user()
        self.playground.create_group_role(
            workspace.scope, ScopeRoles.OWNER, users=[user]
        )
        asset = self.playground.create_signing_key_asset()
        data = asset.data_model
        assert isinstance(data, SigningKeyData)
        response = self.post_asset_create(
            token_key=self.scenario.user_token.key,
            data=self.signing_key_asset_as_dict_request(
                purpose=data.purpose,
                fingerprint=data.fingerprint,
                public_key="PUBLIC KEY",
                workspace=workspace,
            ),
            scope=workspace.scope,
        )
        body = self.assertAPIResponseOk(response)
        self.assertEqual(body["id"], asset.id)

    def assert_asset_equals(
        self, asset: Asset, asset_response: dict[str, Any]
    ) -> None:
        """Assert that asset == asset_response."""
        # Not true in general, but true for all the assets that the view can
        # currently return (since it uses Asset.objects.in_current_scope).
        assert asset.workspace is not None

        self.assertEqual(asset_response["id"], asset.id)
        self.assertEqual(asset_response["data"], asset.data)
        self.assertEqual(asset_response["category"], asset.category)
        self.assertEqual(asset_response["scope"], asset.workspace.scope.name)
        self.assertEqual(asset_response["workspace"], asset.workspace.name)
        self.assertEqual(
            parse_datetime(asset_response["created_at"]), asset.created_at
        )
        self.assertEqual(
            asset_response["work_request"], asset.created_by_work_request_id
        )

    def test_list_asset_by_id(self) -> None:
        """List an asset by ID."""
        workspace = self.playground.get_default_workspace()
        asset = self.playground.create_signing_key_asset()
        response = self.get_asset_list(
            token_key=self.scenario.user_token.key,
            query_params={"asset": asset.id},
            scope=workspace.scope,
        )
        body = self.assertAPIResponseOk(response)
        self.assert_asset_equals(asset, body[0])

    def test_list_asset_by_work_request(self) -> None:
        """List an asset by work request."""
        workspace = self.playground.get_default_workspace()
        work_request = self.playground.create_worker_task()
        asset = self.playground.create_signing_key_asset(
            created_by_work_request=work_request
        )
        response = self.get_asset_list(
            token_key=self.scenario.user_token.key,
            query_params={"work_request": work_request.id},
            scope=workspace.scope,
        )
        body = self.assertAPIResponseOk(response)
        self.assert_asset_equals(asset, body[0])

    def test_list_assets_by_workspace(self) -> None:
        """List assets by workspace."""
        workspace = self.playground.get_default_workspace()
        asset1 = self.playground.create_signing_key_asset()
        asset2 = self.playground.create_signing_key_asset(
            purpose=KeyPurpose.UEFI
        )
        workspace2 = self.playground.create_workspace(name="workspace2")
        self.playground.create_signing_key_asset(workspace=workspace2)
        response = self.get_asset_list(
            token_key=self.scenario.user_token.key,
            query_params={"workspace": workspace.name},
            scope=workspace.scope,
        )
        body = self.assertAPIResponseOk(response)
        self.assertEqual(len(body), 2)
        self.assert_asset_equals(asset1, body[0])
        self.assert_asset_equals(asset2, body[1])

    def test_list_assets_unfiltered(self) -> None:
        """List assets without a filter."""
        workspace = self.playground.get_default_workspace()
        response = self.get_asset_list(
            token_key=self.scenario.user_token.key,
            query_params={},
            scope=workspace.scope,
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        body = response.json()
        self.assertEqual(
            body,
            {
                "title": (
                    '"asset", "work_request", or "workspace" parameters are '
                    'mandatory'
                ),
            },
        )


class AssetPermissionCheckViewTests(TestCase):
    """Tests for AssetPermissionCheckView."""

    playground_memory_file_store = False
    worker: Worker

    @override
    def setUp(self) -> None:
        super().setUp()
        self.worker = self.playground.create_worker(
            worker_type=WorkerType.SIGNING
        )

    def post_permission_check(
        self,
        *,
        asset_category: str,
        asset_slug: str,
        permission_name: str,
        data: dict[str, Any],
        token: Token | None = None,
    ) -> TestResponseType:
        """
        Call self.client.post() to api:asset-permission-check.

        Pass the token.key and content_type="application/json".

        :param token_key: value of the Token header for the request
        :param data: data to post
        """
        if token is None:
            work_request = self.playground.create_worker_task(
                worker=self.worker, mark_running=True
            )
            token = self.playground.create_work_request_token(work_request)
        return self.client.post(
            reverse(
                "api:asset-permission-check",
                kwargs={
                    "asset_category": asset_category,
                    "asset_slug": asset_slug,
                    "permission_name": permission_name,
                },
            ),
            data=data,
            HTTP_TOKEN=token.key,
            content_type="application/json",
        )

    def create_debsign_work_request(
        self,
        asset: Asset,
        upload_artifacts: SourceUploadArtifacts,
        **kwargs: Any,
    ) -> WorkRequest:
        """Create a debsign WorkRequest using asset and upload_artifacts."""
        asset_model = asset.data_model
        assert isinstance(asset_model, SigningKeyData)
        wr = self.playground.create_signing_task(
            task_name="debsign",
            task_data=DebsignData(
                unsigned=upload_artifacts.upload.id,
                key=asset_model.fingerprint,
            ),
            mark_running=True,
            **kwargs,
        )
        wr.dynamic_task_data = DebsignDynamicData(
            unsigned_id=upload_artifacts.upload.id
        ).model_dump()
        wr.save()
        return wr

    def test_permission_check_on_upload(self) -> None:
        workspace = self.playground.get_default_workspace()
        asset = self.playground.create_signing_key_asset()
        upload_artifacts = self.playground.create_upload_artifacts(binary=False)
        work_request = self.create_debsign_work_request(asset, upload_artifacts)
        token = self.playground.create_work_request_token(work_request)
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset.slug,
            permission_name="sign_with",
            token=token,
            data={
                "artifact_id": upload_artifacts.upload.id,
                "work_request_id": work_request.id,
                "workspace": workspace.name,
            },
        )
        body = self.assertAPIResponseOk(response)
        self.assertFalse(body["has_permission"])
        self.assertEqual(body["username"], work_request.created_by.username)
        self.assertEqual(body["user_id"], work_request.created_by.id)
        self.assertEqual(
            body["resource"],
            {
                "architecture": "source",
                "source": "hello",
                "version": "1.0-1",
                "distribution": "unstable",
            },
        )

    def test_permission_check_on_upload_granted(self) -> None:
        workspace = self.playground.get_default_workspace()
        user = self.playground.get_default_user()
        asset = self.playground.create_signing_key_asset()
        asset_usage = self.playground.create_asset_usage(
            resource=asset, workspace=workspace
        )
        self.playground.create_group_role(
            asset_usage, AssetUsageRoles.SIGNER, users=[user]
        )
        upload_artifacts = self.playground.create_upload_artifacts(binary=False)
        work_request = self.create_debsign_work_request(
            asset, upload_artifacts, created_by=user
        )
        token = self.playground.create_work_request_token(work_request)
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset.slug,
            permission_name="sign_with",
            token=token,
            data={
                "artifact_id": upload_artifacts.upload.id,
                "work_request_id": work_request.id,
                "workspace": workspace.name,
            },
        )
        body = self.assertAPIResponseOk(response)
        self.assertTrue(body["has_permission"])
        self.assertEqual(body["username"], user.username)
        self.assertEqual(body["user_id"], work_request.created_by.id)

    def create_sign_work_request(
        self,
        asset: Asset,
        signing_input_artifacts: list[Artifact],
        **kwargs: Any,
    ) -> WorkRequest:
        """Create a sign WorkRequest using asset and signing_input."""
        asset_model = asset.data_model
        assert isinstance(asset_model, SigningKeyData)
        signing_input_ids = [
            signing_input.id for signing_input in signing_input_artifacts
        ]
        wr = self.playground.create_signing_task(
            task_name="sign",
            task_data=SignData(
                purpose=KeyPurpose.UEFI,
                unsigned=LookupMultiple.model_validate(signing_input_ids),
                key=asset_model.fingerprint,
            ),
            mark_running=True,
            **kwargs,
        )
        wr.dynamic_task_data = SignDynamicData(
            unsigned_ids=signing_input_ids
        ).model_dump()
        wr.save()
        return wr

    def test_permission_check_on_signing_input(self) -> None:
        workspace = self.playground.get_default_workspace()
        asset = self.playground.create_signing_key_asset()
        signing_input = self.playground.create_signing_input_artifact()
        work_request = self.create_sign_work_request(asset, [signing_input])
        token = self.playground.create_work_request_token(work_request)

        self.playground.create_artifact_relation(
            signing_input,
            target=self.playground.create_minimal_binary_package_artifact(),
        )

        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset.slug,
            permission_name="sign_with",
            token=token,
            data={
                "artifact_id": signing_input.id,
                "work_request_id": work_request.id,
                "workspace": workspace.name,
            },
        )
        body = self.assertAPIResponseOk(response)
        self.assertFalse(body["has_permission"])
        self.assertEqual(body["username"], work_request.created_by.username)
        self.assertEqual(body["user_id"], work_request.created_by.id)
        self.assertEqual(
            body["resource"],
            {
                "architecture": "all",
                "package": "hello",
                "source": "hello",
                "version": "1.0-1",
            },
        )

    def create_sign_repository_index_work_request(
        self, asset: Asset, suite: Collection, index: Artifact, **kwargs: Any
    ) -> WorkRequest:
        """Create a SignRepositoryIndex work request."""
        asset_model = asset.data_model
        assert isinstance(asset_model, SigningKeyData)
        wr = self.playground.create_signing_task(
            task_name="signrepositoryindex",
            task_data=SignRepositoryIndexData(
                suite_collection=f"{suite.name}@{CollectionCategory.SUITE}",
                unsigned=index.id,
                mode=SigningMode.CLEAR,
                signed_name="InRelease",
            ),
            mark_running=True,
            **kwargs,
        )
        wr.dynamic_task_data = SignRepositoryIndexDynamicData(
            signing_keys=[asset_model.fingerprint], unsigned_id=index.id
        ).model_dump()
        wr.save()
        return wr

    def test_permission_check_on_repository_index(self) -> None:
        workspace = self.playground.get_default_workspace()
        asset = self.playground.create_signing_key_asset()
        AssetUsage.objects.create(asset=asset, workspace=workspace)
        suite = self.playground.create_collection(
            "suite", CollectionCategory.SUITE
        )
        index = self.playground.create_repository_index(
            "Release", skip_add_files_in_store=True
        )
        work_request = self.create_sign_repository_index_work_request(
            asset, suite, index
        )
        token = self.playground.create_work_request_token(work_request)

        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset.slug,
            permission_name="sign_repository_with",
            token=token,
            data={
                "artifact_id": index.id,
                "work_request_id": work_request.id,
                "workspace": workspace.name,
            },
        )
        body = self.assertAPIResponseOk(response)
        self.assertFalse(body["has_permission"])
        self.assertEqual(body["username"], work_request.created_by.username)
        self.assertEqual(body["user_id"], work_request.created_by.id)
        self.assertEqual(body["resource"], {"path": "Release"})

    def test_permission_check_worker_token(self) -> None:
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug="openpgp:ABC123",
            permission_name="sign_with",
            token=self.worker.token,
            data={
                "artifact_id": 1,
                "work_request_id": 1,
                "workspace": "workspace",
            },
        )
        self.assertResponseProblem(
            response,
            "Error",
            detail_pattern="You do not have permission to perform this action.",
            status_code=status.HTTP_403_FORBIDDEN,
        )

    def test_permission_check_unrelated_work_request(self) -> None:
        user = self.playground.get_default_user()
        workspace = self.playground.get_default_workspace()
        asset = self.playground.create_signing_key_asset()
        AssetUsage.objects.create(asset=asset, workspace=workspace)
        artifact = self.playground.create_signing_input_artifact()
        work_request_1 = self.playground.create_worker_task(
            mark_pending=True, worker=self.worker
        )
        work_request_2 = self.playground.create_worker_task(
            mark_running=True, worker=self.worker
        )
        token = self.playground.create_work_request_token(work_request_2)
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset.slug,
            permission_name="sign_with",
            token=token,
            data={
                "artifact_id": artifact.id,
                "work_request_id": work_request_1.id,
                "workspace": workspace.name,
            },
        )
        self.assertResponseProblem(
            response,
            f"{user} cannot display work request {work_request_1.id}",
            status_code=status.HTTP_403_FORBIDDEN,
        )

    def test_permission_check_wrong_worker_type(self) -> None:
        work_request = self.playground.create_worker_task(
            mark_running=True, assign_new_worker=True
        )
        token = self.playground.create_work_request_token(work_request)
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug="openpgp:ABC123",
            permission_name="sign_with",
            token=token,
            data={
                "artifact_id": 1,
                "work_request_id": 1,
                "workspace": "workspace",
            },
        )
        self.assertResponseProblem(
            response,
            "Forbidden",
            detail_pattern=r"Endpoint is not accessible to external workers",
            status_code=status.HTTP_403_FORBIDDEN,
        )

    def test_permission_check_missing_asset(self) -> None:
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug="openpgp:ABC123",
            permission_name="sign_with",
            data={
                "artifact_id": 1,
                "work_request_id": 1,
                "workspace": "workspace",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        body = response.json()
        self.assertEqual(body["title"], "Asset matching query does not exist.")

    def test_permission_check_malformed_slug(self) -> None:
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug="foobar",
            permission_name="sign_with",
            data={
                "artifact_id": 1,
                "work_request_id": 1,
                "workspace": "workspace",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        body = response.json()
        self.assertEqual(
            body["title"],
            (
                "Malformed asset_slug: not enough values to unpack "
                "(expected 2, got 1)"
            ),
        )

    def test_permission_check_missing_artifact(self) -> None:
        asset = self.playground.create_signing_key_asset()
        work_request = self.create_sign_work_request(asset, [])
        token = self.playground.create_work_request_token(work_request)
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset.slug,
            permission_name="sign_with",
            token=token,
            data={
                "artifact_id": 0,
                "work_request_id": work_request.id,
                "workspace": work_request.workspace.name,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        body = response.json()
        self.assertEqual(
            body["title"], "Artifact matching query does not exist."
        )

    def test_permission_check_missing_work_request(self) -> None:
        asset = self.playground.create_signing_key_asset()
        artifact = self.playground.create_signing_input_artifact()
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset.slug,
            permission_name="sign_with",
            data={
                "artifact_id": artifact.id,
                "work_request_id": 0,
                "workspace": artifact.workspace.name,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        body = response.json()
        self.assertEqual(
            body["title"], "WorkRequest matching query does not exist."
        )

    def test_permission_check_work_request_visible(self) -> None:
        # This is an impossible scenario but it's good to have the check for
        # completeness.
        asset = self.playground.create_signing_key_asset()
        artifact = self.playground.create_signing_input_artifact()
        another_scope = self.playground.get_or_create_scope(name="another")
        private_workspace = self.playground.create_workspace(
            scope=another_scope, public=False
        )
        user = self.playground.get_default_user()
        work_request = self.create_sign_work_request(
            asset, [artifact], workspace=private_workspace, created_by=user
        )
        token = self.playground.create_work_request_token(work_request)
        self.assertFalse(
            work_request.can_display(PermissionContext.for_ui_user(user))
        )
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset.slug,
            permission_name="sign_with",
            token=token,
            data={
                "artifact_id": artifact.id,
                "work_request_id": work_request.id,
                "workspace": artifact.workspace.name,
            },
        )
        self.assertResponseProblem(
            response,
            "Workspace not found",
            detail_pattern="Workspace System not found in scope debusine",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_permission_check_artifact_visible(self) -> None:
        asset = self.playground.create_signing_key_asset()
        another_scope = self.playground.get_or_create_scope(name="another")
        # Make the workspace initially public to be able to populate the work
        # request
        private_workspace = self.playground.create_workspace(
            scope=another_scope, public=True
        )
        artifact = self.playground.create_signing_input_artifact(
            workspace=private_workspace
        )
        user = self.playground.get_default_user()
        work_request = self.create_sign_work_request(
            asset, [artifact], created_by=user
        )
        token = self.playground.create_work_request_token(work_request)
        private_workspace.public = False
        private_workspace.save()

        self.assertFalse(
            artifact.can_display(PermissionContext.for_ui_user(user))
        )
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset.slug,
            permission_name="sign_with",
            token=token,
            data={
                "artifact_id": artifact.id,
                "work_request_id": work_request.id,
                "workspace": artifact.workspace.name,
            },
        )
        self.assertResponseProblem(
            response,
            title=f"{user} cannot display artifact {artifact.id}",
            status_code=status.HTTP_403_FORBIDDEN,
        )

    def test_permission_check_artifact_unrelated(self) -> None:
        asset = self.playground.create_signing_key_asset()
        artifact = self.playground.create_signing_input_artifact()
        user = self.playground.get_default_user()
        work_request = self.create_sign_work_request(asset, [], created_by=user)
        token = self.playground.create_work_request_token(work_request)
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset.slug,
            permission_name="sign_with",
            token=token,
            data={
                "artifact_id": artifact.id,
                "work_request_id": work_request.id,
                "workspace": artifact.workspace.name,
            },
        )
        self.assertResponseProblem(
            response,
            f"{user} cannot display artifact {artifact.id}",
            status_code=status.HTTP_403_FORBIDDEN,
        )

    def test_permission_check_wrong_workspace(self) -> None:
        asset = self.playground.create_signing_key_asset()
        artifact = self.playground.create_signing_input_artifact()
        work_request = self.create_sign_work_request(asset, [artifact])
        token = self.playground.create_work_request_token(work_request)
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset.slug,
            permission_name="sign_with",
            token=token,
            data={
                "artifact_id": artifact.id,
                "work_request_id": work_request.id,
                "workspace": "another-workspace",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        body = response.json()
        self.assertEqual(body["title"], "workspace does not match work request")

    def test_permission_check_unknown_permission(self) -> None:
        asset = self.playground.create_signing_key_asset()
        artifact = self.playground.create_signing_input_artifact()
        work_request = self.create_sign_work_request(asset, [artifact])
        token = self.playground.create_work_request_token(work_request)
        response = self.post_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset.slug,
            permission_name="unknown",
            token=token,
            data={
                "artifact_id": artifact.id,
                "work_request_id": work_request.id,
                "workspace": artifact.workspace.name,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        body = response.json()
        self.assertEqual(body["title"], "Unknown permission 'unknown'")

    def test_describe_artifact_repository_index(self) -> None:
        artifact = self.playground.create_repository_index("Release")
        result = AssetPermissionCheckView.describe_artifact(artifact)
        self.assertEqual(result, {"path": "Release"})

    def test_describe_artifact_signing_input_standalone(self) -> None:
        artifact = self.playground.create_signing_input_artifact()
        with self.assertLogsContains(
            (
                f"INFO:debusine.server.views.assets:Unable to fully describe "
                f"artifact {artifact!r}, no related debian:binary-package."
            ),
            logger=log,
        ):
            result = AssetPermissionCheckView.describe_artifact(artifact)
            self.assertEqual(result, {"package": "hello"})

    def test_describe_artifact_signing_input_multiple_relations(self) -> None:
        artifact = self.playground.create_signing_input_artifact()
        for _ in range(2):
            self.playground.create_artifact_relation(
                artifact,
                target=self.playground.create_minimal_binary_package_artifact(),
            )
        with self.assertLogsContains(
            (
                f"INFO:debusine.server.views.assets:Unable to fully describe "
                f"artifact {artifact!r}, multiple related "
                f"debian:binary-package."
            ),
            logger=log,
        ):
            result = AssetPermissionCheckView.describe_artifact(artifact)
            self.assertEqual(result, {"package": "hello"})

    def test_describe_artifact_signing_input_without_name(self) -> None:
        """Looks up the package from the debian:binary-package, if missing."""
        artifact = self.playground.create_signing_input_artifact(
            binary_package_name=None
        )
        self.playground.create_artifact_relation(
            artifact,
            target=self.playground.create_minimal_binary_package_artifact(),
        )
        result = AssetPermissionCheckView.describe_artifact(artifact)
        self.assertEqual(result["package"], "hello")

    def test_describe_artifact_upload_unknown(self) -> None:
        upload_artifacts = self.playground.create_upload_artifacts(binary=False)
        artifact = upload_artifacts.upload
        artifact.data["type"] = "unknown"
        with self.assertRaisesRegex(
            NotImplementedError,
            f"Unable to describe upload artifact {artifact}",
        ):
            AssetPermissionCheckView.describe_artifact(artifact)

    def test_describe_artifact_unknown(self) -> None:
        artifact = self.playground.create_minimal_binary_package_artifact()
        with self.assertRaisesRegex(
            NotImplementedError,
            f"Unable to describe artifact {artifact}",
        ):
            AssetPermissionCheckView.describe_artifact(artifact)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests the artifact UI helper."""

from typing import ClassVar, override

from django.core.exceptions import ImproperlyConfigured
from django.test import RequestFactory
from django.urls import reverse

from debusine.db.context import context
from debusine.db.models import Group
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views.places import ResourcePlace
from debusine.web.views.ui.auth import GroupUI


class GroupUITests(TestCase):
    """Test the :py:class:`GroupUI` class."""

    scenario = scenarios.DefaultContext()

    group: ClassVar[Group]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.group = cls.playground.create_group("test")

    def get_helper(self, group: Group) -> GroupUI:
        """Instantiate a UI helper for a group."""
        assert context.user
        factory = RequestFactory()
        request = factory.get("/")
        request.user = context.user
        return GroupUI(request, group)

    def test_place_unauthenticated(self) -> None:
        context.set_permission_context_scoped(
            self.scenario.scope, PermissionContext.for_ui_anonymous()
        )
        with self.assertRaisesRegex(
            ImproperlyConfigured, "GroupUI.place invoked by an anonymous user"
        ):
            self.get_helper(self.group).place

    def test_place_authenticated(self) -> None:
        self.scenario.set_current()
        place = self.get_helper(self.group).place
        assert isinstance(place, ResourcePlace)
        self.assertEqual(place.title, f"Group {self.group}")
        self.assertEqual(place.breadcrumb, self.group.name)
        self.assertEqual(place.url, self.group.get_absolute_url())

        assert place.parent is not None
        self.assertEqual(
            place.parent.url,
            reverse(
                "user:groups", kwargs={"username": self.scenario.user.username}
            ),
        )
        assert place.parent.parent is not None
        self.assertEqual(place.parent.parent.url, reverse("scopes:detail"))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the auth views."""

import datetime as dt
from textwrap import dedent
from typing import Any, ClassVar, override
from urllib.parse import urlencode

import lxml
from django.conf import settings
from django.contrib.auth import get_user_model
from django.template.response import SimpleTemplateResponse
from django.test import RequestFactory, override_settings
from django.urls import reverse
from django.utils import timezone
from rest_framework import status

from debusine.db.context import context
from debusine.db.models import Group, Token, User, auth
from debusine.db.models.auth import GroupAuditLog, GroupMembership, Identity
from debusine.db.playground import scenarios
from debusine.server.signon.providers import Provider
from debusine.test.django import (
    AllowAll,
    DenyAll,
    TestCase,
    TestResponseType,
    override_permission,
)
from debusine.web.views.auth import UserDetailView
from debusine.web.views.tests.utils import (
    ViewTestMixin,
    date_format,
    html_check_icon,
)


class LoginViewTests(TestCase):
    """Tests for the LoginView class."""

    def test_get(self) -> None:
        """Test plain get, no next= argument."""
        response = self.client.get(reverse("login"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(self.client.session["sso_callback_next_url"], "/")

    def test_get_valid_next(self) -> None:
        """Test plain get, valid next= argument."""
        response = self.client.get(reverse("login") + "?next=/foo")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(self.client.session["sso_callback_next_url"], "/foo")

    def test_get_invalid_next(self) -> None:
        """Test plain get, invalid next= argument."""
        response = self.client.get(
            reverse("login") + "?next=https://example.com"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(self.client.session["sso_callback_next_url"], "/")

    def test_successful_login(self) -> None:
        """Successful login redirect to the homepage with authenticated user."""
        username = "testuser"
        password = "testpassword"

        get_user_model().objects.create_user(
            username=username, password=password
        )

        response = self.client.post(
            reverse("login"),
            data={"username": username, "password": password},
        )

        self.assertRedirects(response, reverse("homepage:homepage"))
        self.assertTrue(response.wsgi_request.user.is_authenticated)

    def test_nexturl_in_session(self) -> None:
        """Login leaves the next url in the session."""
        response = self.client.get(
            reverse("login") + "?next=foo",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(self.client.session["sso_callback_next_url"], "foo")


class LogoutViewTests(TestCase):
    """Tests for the logout view."""

    def test_successful_render(self) -> None:
        """
        Logout template is rendered.

        The view is implemented (and tested) as part of Django. The template
        is implemented by Debusine, and it's rendered here to assert that
        no missing template variables and contains the expected strings.
        """
        response = self.client.post(reverse("logout"))

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class UserDetailViewTests(ViewTestMixin, TestCase):
    """Test for UserDetailView class."""

    scenario = scenarios.DefaultScopeUser()

    def test_get_place_full_name(self) -> None:
        view = UserDetailView()
        view.object = self.scenario.user
        view.request = RequestFactory().get("/")
        place = view.get_place()
        self.assertEqual(place.title, self.scenario.user.get_full_name())
        self.assertEqual(place.url, self.scenario.user.get_absolute_url())
        self.assertIsNone(place.icon)
        self.assertEqual(place.description, "User Play Ground <playground>")
        self.assertIsNone(place.parent)

    def test_get_place_username(self) -> None:
        # Full name is not set
        user = self.playground.create_user(username="John")
        view = UserDetailView()
        view.object = user
        view.request = RequestFactory().get("/")
        place = view.get_place()
        self.assertEqual(place.title, user.username)
        self.assertEqual(place.url, user.get_absolute_url())
        self.assertIsNone(place.icon)
        self.assertEqual(place.description, "User <John>")
        self.assertIsNone(place.parent)

    def test_anonymous(self) -> None:
        """Test accessing as anonymous user."""
        response = self.client.get(self.scenario.user.get_absolute_url())
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(
            response.context["exception"], "No User matches the given query."
        )

    def test_detail(self) -> None:
        """Test viewing the default user."""
        self.client.force_login(self.scenario.user)
        response = self.client.get(self.scenario.user.get_absolute_url())
        tree = self.assertResponseHTML(response)

        dl = self.assertHasElement(tree, "//dl[@id='person-info']")
        self.assertTextContentEqual(
            dl.dd[0], self.scenario.user.get_full_name()
        )
        self.assertTextContentEqual(dl.dd[1], self.scenario.user.username)
        self.assertTextContentEqual(
            dl.dd[2], str(self.scenario.user.token_set.count())
        )
        self.assertEqual(
            dl.dd[2].a.attrib["href"],
            reverse(
                "user:token-list",
                kwargs={"username": self.scenario.user.username},
            ),
        )

        self.assertFalse(tree.xpath("//table[@id='person-groups']"))
        self.assertFalse(tree.xpath("//table[@id='person-identities']"))

    def test_detail_groups(self) -> None:
        """Test viewing a user with groups."""
        self.client.force_login(self.scenario.user)
        debusine_groups = ("foo", "bar", "baz")
        another_groups = ("wobble", "wibble", "wabble")

        for name in debusine_groups:
            self.playground.create_group(
                name=name, scope=self.scenario.scope, users=[self.scenario.user]
            )

        scope = self.playground.get_or_create_scope("another")
        for name in another_groups:
            self.playground.create_group(
                name=name, scope=scope, users=[self.scenario.user]
            )

        response = self.client.get(self.scenario.user.get_absolute_url())
        tree = self.assertResponseHTML(response)

        table = self.assertHasElement(tree, "//table[@id='person-groups']")
        tbody = self.assertHasElement(table, "tbody")
        self.assertTextContentEqual(tbody.tr[0].td[0], "another")
        for li, name in zip(tbody.tr[0].td[1].ul.li, sorted(another_groups)):
            self.assertTextContentEqual(li, name)
        self.assertTextContentEqual(tbody.tr[1].td[0], "debusine")
        for li, name in zip(tbody.tr[1].td[1].ul.li, sorted(debusine_groups)):
            self.assertTextContentEqual(li, name)

    @override_settings(SIGNON_PROVIDERS=[Provider(name="salsa", label="Salsa")])
    def test_detail_identities(self) -> None:
        """Test viewing a user with SSO identities."""
        self.client.force_login(self.scenario.user)

        identity = Identity.objects.create(
            user=self.scenario.user,
            issuer="salsa",
            subject="playground@debian.org",
        )

        response = self.client.get(self.scenario.user.get_absolute_url())
        tree = self.assertResponseHTML(response)

        table = self.assertHasElement(tree, "//table[@id='person-identities']")
        tr = self.assertHasElement(table, "tbody/tr")
        self.assertTextContentEqual(tr.td[0], "Salsa")
        self.assertTextContentEqual(tr.td[1], "playground@debian.org")
        self.assertTextContentEqual(
            tr.td[2], identity.last_used.strftime("%Y-%m-%d %H:%M")
        )

    @override_settings(SIGNON_PROVIDERS=[Provider(name="salsa", label="Salsa")])
    def test_detail_identities_cannot_manage(self) -> None:
        """Test viewing another user with SSO identities."""
        self.client.force_login(self.scenario.user)

        user = self.playground.create_user("test")
        Identity.objects.create(
            user=user,
            issuer="salsa",
            subject="playground@debian.org",
        )

        response = self.client.get(user.get_absolute_url())
        tree = self.assertResponseHTML(response)
        self.assertFalse(tree.xpath("//table[@id='person-identities']"))

    def test_detail_identities_without_provider(self) -> None:
        """Test showing a user with identities without a SSO provider."""
        self.client.force_login(self.scenario.user)

        Identity.objects.create(
            user=self.scenario.user,
            issuer="salsa",
            subject="playground@debian.org",
        )
        response = self.client.get(self.scenario.user.get_absolute_url())
        tree = self.assertResponseHTML(response)
        self.assertFalse(tree.xpath("//table[@id='person-identities']"))

    @override_settings(SIGNON_PROVIDERS=[Provider(name="salsa", label="Salsa")])
    def test_detail_identities_with_wrong_provider(self) -> None:
        """Test showing a user with identities of nonexisting SSO provider."""
        self.client.force_login(self.scenario.user)

        Identity.objects.create(
            user=self.scenario.user,
            issuer="nachos",
            subject="playground@debian.org",
        )
        response = self.client.get(self.scenario.user.get_absolute_url())
        tree = self.assertResponseHTML(response)
        self.assertFalse(tree.xpath("//table[@id='person-identities']"))


class UserIDRedirectView(TestCase):
    """Tests for UserIDRedirectView."""

    scenario = scenarios.DefaultScopeUser()

    def test_anonymous(self) -> None:
        response = self.client.get(
            self.scenario.user.get_absolute_user_id_url()
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(
            response.context["exception"], "No user found matching the query"
        )

    def test_redirect(self) -> None:
        self.client.force_login(self.scenario.user)
        response = self.client.get(
            self.scenario.user.get_absolute_user_id_url()
        )
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertRedirects(response, self.scenario.user.get_absolute_url())


class UserTokenListViewTests(ViewTestMixin, TestCase):
    """Tests for UserTokenListView class."""

    scenario = scenarios.DefaultScopeUser()

    def test_only_one_scope(self) -> None:
        """Test hiding the scopes list if there is only one."""
        self.client.force_login(self.scenario.user)
        response = self.client.get(
            reverse(
                "user:token-list",
                kwargs={"username": self.scenario.user.username},
            )
        )
        tree = self.assertResponseHTML(response)
        sample_cfg = self.assertHasElement(tree, "//pre[@id='example-config']")
        fqdn = settings.DEBUSINE_FQDN
        self.assertTextContentEqual(
            sample_cfg,
            dedent(
                f"""
                [General]
                default-server = {fqdn}

                [server:{fqdn}]
                api-url = https://{fqdn}/api
                scope = {self.scenario.scope}
                token = generated-token-goes-here
                """
            ),
        )
        self.assertFalse(tree.xpath("ul[@id='scope-list']"))

    def test_two_scopes(self) -> None:
        """Test when there is more than one scope."""
        self.playground.get_or_create_scope("test")
        self.client.force_login(self.scenario.user)
        response = self.client.get(
            reverse(
                "user:token-list",
                kwargs={"username": self.scenario.user.username},
            )
        )
        tree = self.assertResponseHTML(response)
        sample_cfg = self.assertHasElement(tree, "//pre[@id='example-config']")
        fqdn = settings.DEBUSINE_FQDN
        scope_list = " or ".join(str(s) for s in response.context["scopes"])
        self.assertTextContentEqual(
            sample_cfg,
            dedent(
                f"""
                [General]
                default-server = {fqdn}

                [server:{fqdn}]
                api-url = https://{fqdn}/api
                scope = {scope_list}
                token = generated-token-goes-here
                """
            ),
        )

    def test_logged_no_tokens(self) -> None:
        """Logged in user with no tokens."""
        another_user = get_user_model().objects.create_user(
            username="notme", password="testpassword", email="test2@example.com"
        )
        another_token = self.playground.create_user_token(user=another_user)

        self.client.force_login(self.scenario.user)
        response = self.client.get(
            reverse(
                "user:token-list",
                kwargs={"username": self.scenario.user.username},
            )
        )
        tree = self.assertResponseHTML(response)
        self.assertNavHasUser(tree, self.scenario.user)

        self.assertContains(
            response,
            "You do not currently have any tokens",
        )

        create_token_url = reverse(
            "user:token-create",
            kwargs={"username": self.scenario.user.username},
        )
        self.assertContains(
            response,
            f'<a href="{create_token_url}">create a new token</a>',
            html=True,
        )

        # Ensure that a token for another user is not displayed
        self.assertNotContains(response, another_token.key)

    def row_for_token(self, token: Token) -> str:
        """Return HTML for the row of the table for token."""
        edit_url = reverse(
            "user:token-edit",
            kwargs={"pk": token.pk, "username": self.scenario.user.username},
        )
        delete_url = reverse(
            "user:token-delete",
            kwargs={"pk": token.pk, "username": self.scenario.user.username},
        )

        if token.last_seen_at is not None:
            last_seen_at = date_format(token.last_seen_at)
        else:
            last_seen_at = "Never"

        return (
            "<tr>"
            f"<td>{html_check_icon(token.enabled)}</td>"
            f"<td>{date_format(token.created_at)}</td>"
            f"<td>{last_seen_at}</td>"
            f"<td>{token.comment}</td>"
            f'<td><a href="{delete_url}">Delete</a>&nbsp;|&nbsp;'
            f'<a href="{edit_url}">Edit</a></td>'
            "</tr>"
        )

    def test_logged_two_tokens(self) -> None:
        """
        Logged user with two tokens: both are displayed.

        Assert tokens are ordered by created_at.
        """
        token1 = self.playground.create_user_token(
            user=self.scenario.user,
            comment="Token 1",
            last_seen_at=timezone.now(),
        )
        token2 = self.playground.create_user_token(
            user=self.scenario.user, comment="Token 2"
        )

        self.client.force_login(self.scenario.user)
        response = self.client.get(
            reverse(
                "user:token-list",
                kwargs={"username": self.scenario.user.username},
            )
        )

        self.assertContains(response, self.row_for_token(token1), html=True)
        self.assertContains(response, self.row_for_token(token2), html=True)

        # Check tokens are ordered_by created_

        # token1 is displayed earlier than token2 (was created before)
        self.assertLess(
            response.content.index(token1.comment.encode("utf-8")),
            response.content.index(token2.comment.encode("utf-8")),
        )

        # Make token1 to be created_at after token2 created_at
        token1.created_at = token2.created_at + dt.timedelta(minutes=1)
        token1.save()

        response = self.client.get(
            reverse(
                "user:token-list",
                kwargs={"username": self.scenario.user.username},
            )
        )

        # Tokens appear in the opposite order than before
        self.assertGreater(
            response.content.index(token1.comment.encode("utf-8")),
            response.content.index(token2.comment.encode("utf-8")),
        )

    def test_no_logged(self) -> None:
        """Request with a non-logged user: does not return any token."""
        response = self.client.get(
            reverse(
                "user:token-list",
                kwargs={"username": self.scenario.user.username},
            )
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(
            response.context["exception"], "No User matches the given query."
        )


class UserTokenCreateViewTests(ViewTestMixin, TestCase):
    """Tests for UserTokenCreateView."""

    user: ClassVar[User]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user = get_user_model().objects.create_user(
            username="testuser", password="testpass"
        )

    def test_create_token(self) -> None:
        """Post to "user:token-create" to create a token."""
        self.client.force_login(self.user)
        comment = "Test 1"
        response = self.client.post(
            reverse(
                "user:token-create",
                kwargs={"username": self.user.username},
            ),
            {"comment": comment, "enabled": "true"},
        )
        # This post, even if successful, renders a template and does not
        # redirect
        tree = self.assertResponseHTML(response)

        # The token got created, enabled, with user assigned
        token = Token.objects.get(comment=comment)
        self.assertEqual(token.user, self.user)
        self.assertTrue(token.enabled)

        # No expiration, no use yet
        self.assertIsNone(token.expire_at)
        self.assertIsNone(token.last_seen_at)

        # Ensure token key shown matches the one of the new token
        key_element = self.assertHasElement(tree, "//code[@id='token-key']")
        key = self.get_node_text_normalized(key_element)
        self.assertEqual(token, Token.objects.get_token_or_none(key))

    def test_no_logged(self) -> None:
        """Request with a non-logged user: cannot create a token."""
        response = self.client.get(
            reverse(
                "user:token-create",
                kwargs={"username": self.user.username},
            )
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(
            response.context["exception"], "No User matches the given query."
        )

    def test_initial_form(self) -> None:
        """Get request to ensure the form is displayed."""
        self.client.force_login(self.user)

        response = self.client.get(
            reverse(
                "user:token-create",
                kwargs={"username": self.user.username},
            )
        )
        tree = self.assertResponseHTML(response)
        title = self.assertHasElement(tree, "//title")
        self.assertTextContentEqual(title, "debusine - testuser - Create token")
        h1 = self.assertHasElement(tree, "//h1")
        self.assertTextContentEqual(h1, "Create token")
        self.assertContains(
            response,
            '<input class="btn btn-primary btn-sm" '
            'type="submit" value="Create">',
            html=True,
        )


class UserTokenUpdateViewTests(TestCase):
    """Tests for UserTokenCreateView when editing a token."""

    scenario = scenarios.DefaultScopeUserAPI()

    def test_edit_token(self) -> None:
        """Get "user:token-edit" to edit a token."""
        token = self.scenario.user_token
        self.client.force_login(self.scenario.user)

        enabled = True
        comment = "This is a token"

        response = self.client.post(
            reverse(
                "user:token-edit",
                kwargs={
                    "pk": token.pk,
                    "username": self.scenario.user.username,
                },
            ),
            {"comment": comment, "enabled": enabled},
        )

        self.assertRedirects(
            response,
            reverse(
                "user:token-list",
                kwargs={"username": self.scenario.user.username},
            ),
        )

        token.refresh_from_db()

        self.assertEqual(token.comment, comment)
        self.assertEqual(token.enabled, enabled)

    def test_get_404_not_found(self) -> None:
        """Get of token/PK/ that exist but belongs to another user: 404."""
        token = self.scenario.user_token
        user2 = get_user_model().objects.create(
            username="testuser2",
            password="testpass2",
            email="user2@example.com",
        )

        self.client.force_login(user2)
        response = self.client.get(
            reverse(
                "user:token-edit",
                kwargs={
                    "pk": token.pk,
                    "username": self.scenario.user.username,
                },
            )
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(
            response.context["exception"],
            "testuser2 cannot manage user playground",
        )
        self.assertNotContains(
            response, token.hash, status_code=status.HTTP_403_FORBIDDEN
        )

    def test_put_404_not_found(self) -> None:
        """Put of token/PK/ that exist but belongs to another user: 404."""
        token = self.scenario.user_token
        user2 = get_user_model().objects.create(
            username="testuser2",
            password="testpass2",
            email="user2@example.com",
        )

        self.client.force_login(user2)
        comment = "This is the new comment"
        enabled = False
        response = self.client.put(
            reverse(
                "user:token-edit",
                kwargs={
                    "pk": token.pk,
                    "username": self.scenario.user.username,
                },
            ),
            {"comment": comment, "enabled": enabled},
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(
            response.context["exception"],
            "testuser2 cannot manage user playground",
        )
        self.assertNotContains(
            response, token.hash, status_code=status.HTTP_403_FORBIDDEN
        )

    def test_initial_form(self) -> None:
        """Get request to ensure the form is displayed."""
        token = self.scenario.user_token
        self.client.force_login(self.scenario.user)

        response = self.client.get(
            reverse(
                "user:token-edit",
                kwargs={
                    "pk": token.pk,
                    "username": self.scenario.user.username,
                },
            )
        )
        tree = self.assertResponseHTML(response)
        title = self.assertHasElement(tree, "//title")
        self.assertTextContentEqual(title, "debusine - playground - Edit token")
        h1 = self.assertHasElement(tree, "//h1")
        self.assertTextContentEqual(h1, "Edit token")

        self.assertContains(
            response,
            '<input class="btn btn-primary btn-sm" type="submit" value="Edit">',
            html=True,
        )
        assert isinstance(response, SimpleTemplateResponse)
        assert response.context_data is not None
        self.assertEqual(
            response.context_data["form"]["comment"].initial,
            token.comment,
        )


class UserTokenDeleteViewTests(TestCase):
    """Tests for UserTokenDeleteView."""

    scenario = scenarios.DefaultScopeUserAPI()

    def test_get_delete_token_authenticated_user(self) -> None:
        """Get request to delete a token. Token information is returned."""
        token = self.scenario.user_token
        token.last_seen_at = timezone.now()
        token.save()
        self.client.force_login(self.scenario.user)

        response = self.client.get(
            reverse(
                "user:token-delete",
                kwargs={
                    "pk": token.pk,
                    "username": self.scenario.user.username,
                },
            )
        )

        expected_contents = (
            "<ul>"
            f"<li>Comment: {token.comment}</li>"
            f"<li>Enabled: {html_check_icon(token.enabled)}</li>"
            f"<li>Created at: {date_format(token.created_at)}</li>"
            f"<li>Last seen at: {date_format(token.last_seen_at)}</li>"
            "</ul>"
        )

        self.assertContains(response, expected_contents, html=True)

    def test_post_delete_token_authenticated_user(self) -> None:
        """Post request to delete a token. The token is deleted."""
        token = self.scenario.user_token
        self.client.force_login(self.scenario.user)

        response = self.client.post(
            reverse(
                "user:token-delete",
                kwargs={
                    "pk": token.pk,
                    "username": self.scenario.user.username,
                },
            )
        )

        self.assertRedirects(
            response,
            reverse(
                "user:token-list",
                kwargs={"username": self.scenario.user.username},
            ),
        )

        # The token was deleted
        self.assertEqual(Token.objects.filter(pk=token.pk).count(), 0)

    def test_get_delete_token_non_authenticated(self) -> None:
        """
        Non authenticated Get and Post request. Return 403.

        The token is not displayed.
        """
        token = self.scenario.user_token
        response = self.client.get(
            reverse(
                "user:token-delete",
                kwargs={
                    "pk": token.pk,
                    "username": self.scenario.user.username,
                },
            )
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(
            response.context["exception"], "No User matches the given query."
        )
        self.assertNotContains(
            response, token.hash, status_code=status.HTTP_404_NOT_FOUND
        )

        response = self.client.post(
            reverse(
                "user:token-delete",
                kwargs={
                    "pk": token.pk,
                    "username": self.scenario.user.username,
                },
            )
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(
            response.context["exception"], "No User matches the given query."
        )
        self.assertNotContains(
            response, token.hash, status_code=status.HTTP_404_NOT_FOUND
        )

    def test_get_post_delete_token_another_user(self) -> None:
        """
        Get and Post request delete token from another user. Return 404.

        The token is not deleted neither displayed.
        """
        token = self.scenario.user_token
        user2 = get_user_model().objects.create_user(
            username="testuser2",
            password="testpass2",
            email="user2@example.com",
        )

        self.client.force_login(user2)

        response = self.client.get(
            reverse(
                "user:token-delete",
                kwargs={
                    "pk": token.pk,
                    "username": self.scenario.user.username,
                },
            )
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(
            response.context["exception"],
            "testuser2 cannot manage user playground",
        )
        self.assertNotContains(
            response, token.hash, status_code=status.HTTP_403_FORBIDDEN
        )

        response = self.client.post(
            reverse(
                "user:token-delete",
                kwargs={
                    "pk": token.pk,
                    "username": self.scenario.user.username,
                },
            )
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(
            response.context["exception"],
            "testuser2 cannot manage user playground",
        )
        self.assertNotContains(
            response, token.hash, status_code=status.HTTP_403_FORBIDDEN
        )

        self.assertEqual(Token.objects.filter(pk=token.pk).count(), 1)


class UserGroupListViewTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`UserGroupListView`."""

    scenario = scenarios.DefaultContext()
    group: ClassVar[Group]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.group = cls.playground.create_group("testgroup")

    def assertGetSucceeds(
        self, user: User, visitor: User | None = None
    ) -> TestResponseType:
        """GET the page and ensure the request succeeds."""
        if visitor is None:
            visitor = user
        self.client.force_login(visitor)
        response = self.client.get(
            reverse("user:groups", kwargs={"username": user.username})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        return response

    def test_get_self_no_groups(self) -> None:
        """Get one's own empty group list."""
        response = self.assertGetSucceeds(self.scenario.user)
        tree = self.assertResponseHTML(response)
        intro = self.assertHasElement(tree, "//p[@id='no-groups-message']")
        self.assertTextContentEqual(
            intro, "You are not currently a member of any group."
        )

    def test_get_visitor_no_groups(self) -> None:
        """Get another person's empty group list."""
        manager = User.objects.create_user("manager")
        with override_permission(User, "can_manage", AllowAll):
            response = self.assertGetSucceeds(
                self.scenario.user, visitor=manager
            )
        tree = self.assertResponseHTML(response)
        intro = self.assertHasElement(tree, "//p[@id='no-groups-message']")
        self.assertTextContentEqual(
            intro,
            f"{self.scenario.user.get_full_name()}"
            f" <{self.scenario.user.username}>"
            " is not currently a member of any group.",
        )

    def test_get_self(self) -> None:
        """Get one's own group list."""
        self.playground.add_user(self.group, self.scenario.user)
        response = self.assertGetSucceeds(self.scenario.user)
        tree = self.assertResponseHTML(response)
        intro = self.assertHasElement(tree, "//p[@id='intro']")
        self.assertTextContentEqual(
            intro, "This page lists all groups of which you are a member."
        )
        table = self.assertHasElement(tree, "//table[@id='group-list']")
        self.assertEqual(len(table.tbody.tr), 1)
        tr = table.tbody.tr[0]
        self.assertTextContentEqual(tr.td[0].a, self.group.name)
        self.assertEqual(tr.td[0].a.get("href"), self.group.get_absolute_url())
        self.assertTextContentEqual(tr.td[1], "1")
        self.assertTextContentEqual(tr.td[2], "no")
        self.assertTextContentEqual(tr.td[3], "member")

    def test_get_self_hidden_groups(self) -> None:
        """Get one's own group list."""
        self.playground.add_user(self.group, self.scenario.user)
        with override_permission(Group, "can_display", DenyAll):
            response = self.assertGetSucceeds(self.scenario.user)
        tree = self.assertResponseHTML(response)
        self.assertHasElement(tree, "//p[@id='no-groups-message']")

    def test_get_visitor(self) -> None:
        """Get another person's group list."""
        self.playground.add_user(self.group, self.scenario.user)
        manager = User.objects.create_user("manager")
        with override_permission(User, "can_manage", AllowAll):
            response = self.assertGetSucceeds(
                self.scenario.user, visitor=manager
            )
        tree = self.assertResponseHTML(response)
        intro = self.assertHasElement(tree, "//p[@id='intro']")
        self.assertTextContentEqual(
            intro,
            "This page lists all groups of which"
            f" {self.scenario.user.get_full_name()} "
            f" <{self.scenario.user.username}> is a member.",
        )
        table = self.assertHasElement(tree, "//table[@id='group-list']")
        self.assertEqual(len(table.tbody.tr), 1)
        tr = table.tbody.tr[0]
        self.assertTextContentEqual(tr.td[0].a, self.group.name)
        self.assertEqual(tr.td[0].a.get("href"), self.group.get_absolute_url())
        self.assertTextContentEqual(tr.td[1], "1")
        self.assertTextContentEqual(tr.td[2], "no")
        self.assertTextContentEqual(tr.td[3], "member")


class GroupDetailViewTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`GroupDetailView`."""

    scenario = scenarios.DefaultContext()
    group: ClassVar[Group]
    other_user: ClassVar[User]
    group_queried: Group | None

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.group = cls.playground.create_group("testgroup")
        cls.other_user = cls.playground.create_user("other")

    @override
    def setUp(self) -> None:
        """Reset the group_queried cache."""
        super().setUp()
        self.group_queried = None

    def assertGetSucceeds(self, group: Group) -> TestResponseType:
        """GET the page and ensure the request succeeds."""
        self.group_queried = group
        self.client.force_login(self.scenario.user)
        response = self.client.get(group.get_absolute_url())
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        return response

    def assertEphemeral(
        self,
        tree: lxml.objectify.ObjectifiedElement,
        expected: bool = True,
    ) -> lxml.objectify.ObjectifiedElement | None:
        """Check that the page reports the group as ephemeral."""
        path = "//p[@id='info-ephemeral']"
        if expected:
            return self.assertHasElement(tree, path)
        else:
            self.assertFalse(tree.xpath(path))
            return None

    def assertFormErrors(
        self,
        tree: lxml.objectify.ObjectifiedElement,
        expected: bool = True,
    ) -> lxml.objectify.ObjectifiedElement | None:
        path = "//ul[@class='errorlist']"
        if expected:
            return self.assertHasElement(tree, path)
        else:
            self.assertFalse(tree.xpath(path))
            return None

    def assertGroupTable(
        self,
        tree: lxml.objectify.ObjectifiedElement,
        contents: list[tuple[User, Group.Roles]],
        is_admin: bool = False,
    ) -> None:
        """Check the contents of the group members table."""
        assert self.group_queried
        members = self.assertHasElement(tree, "//table[@id='members']")
        self.assertEqual(len(members.tbody.tr), len(contents))
        for idx, (user, role) in enumerate(contents):
            gm = GroupMembership.objects.get(
                group=self.group_queried, user=user
            )
            tr = members.tbody.tr[idx]
            suffix = " (you)" if user == self.scenario.user else ""
            self.assertTextContentEqual(tr.td[0], user.username + suffix)
            self.assertTextContentEqual(tr.td[1], user.get_full_name())
            self.assertTextContentEqual(tr.td[2], role)
            if is_admin:
                self.assertEqual(
                    tr.td[2].a.get("href"), gm.get_absolute_url_update()
                )

                self.assertTextContentEqual(tr.td[3].a, "Remove")
                self.assertEqual(
                    tr.td[3].a.get("href"), gm.get_absolute_url_remove()
                )
            else:
                self.assertEqual(len(tr.td), 3)
        if is_admin:
            self.assertHasElement(tree, "//form[@id='add-user-form']")
        else:
            self.assertFalse(tree.xpath("//form[@id='add-user-form']"))

    def test_permissions(self) -> None:
        """Test that the right permissions are enforced."""
        url = self.group.get_absolute_url()
        self.client.force_login(self.scenario.user)

        self.assertEnforcesPermission(
            self.group.can_display,
            url,
            "get_context_data",
            method="get",
            error_code=status.HTTP_404_NOT_FOUND,
        )

        self.assertEnforcesPermission(
            self.group.can_manage,
            url,
            "form_valid",
            method="post",
            data={"username": "other"},
        )

    def test_get_as_anonymous(self) -> None:
        """Test that the right permissions are enforced."""
        response = self.client.get(self.group.get_absolute_url())
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_get_as_nonmember(self) -> None:
        """View the page as a group member."""
        self.playground.add_user(self.group, self.other_user)
        response = self.assertGetSucceeds(self.group)
        tree = self.assertResponseHTML(response)
        info_role = self.assertHasElement(tree, "//p[@id='info-role']")
        self.assertTextContentEqual(
            info_role, "You are not a member of this group."
        )
        self.assertEphemeral(tree, False)
        self.assertFormErrors(tree, False)
        self.assertGroupTable(tree, [(self.other_user, Group.Roles.MEMBER)])

    def test_get_ephemeral(self) -> None:
        """View the page as a group member."""
        with context.disable_permission_checks():
            self.group.ephemeral = True
            self.group.save()
        response = self.assertGetSucceeds(self.group)
        tree = self.assertResponseHTML(response)
        self.assertEphemeral(tree, True)

    def test_get_as_member(self) -> None:
        """View the page as a group member."""
        self.playground.add_user(self.group, self.scenario.user)
        self.playground.add_user(self.group, self.other_user)
        response = self.assertGetSucceeds(self.group)
        tree = self.assertResponseHTML(response)
        info_role = self.assertHasElement(tree, "//p[@id='info-role']")
        self.assertTextContentEqual(info_role, "You have role member.")
        self.assertEphemeral(tree, False)
        self.assertFormErrors(tree, False)
        self.assertGroupTable(
            tree,
            [
                (self.other_user, Group.Roles.MEMBER),
                (self.scenario.user, Group.Roles.MEMBER),
            ],
        )

    def test_get_as_admin(self) -> None:
        """View the page as a group member."""
        self.playground.add_user(
            self.group, self.scenario.user, Group.Roles.ADMIN
        )
        self.playground.add_user(self.group, self.other_user)
        response = self.assertGetSucceeds(self.group)
        tree = self.assertResponseHTML(response)
        info_role = self.assertHasElement(tree, "//p[@id='info-role']")
        self.assertTextContentEqual(info_role, "You have role admin.")
        self.assertEphemeral(tree, False)
        self.assertFormErrors(tree, False)
        self.assertGroupTable(
            tree,
            [
                (self.scenario.user, Group.Roles.ADMIN),
                (self.other_user, Group.Roles.MEMBER),
            ],
            is_admin=True,
        )

    def test_post_add_user(self) -> None:
        """Add a user to the group."""
        self.playground.add_user(
            self.group, self.scenario.user, Group.Roles.ADMIN
        )
        self.client.force_login(self.scenario.user)
        response = self.client.post(
            self.group.get_absolute_url(),
            data={"username": self.other_user.username},
        )
        self.assertRedirects(response, self.group.get_absolute_url())
        membership = GroupMembership.objects.get(
            group=self.group, user=self.other_user
        )
        self.assertEqual(membership.role, Group.Roles.MEMBER)

    def test_post_add_user_validation_failed(self) -> None:
        """Adding a user to the group failed validation."""
        self.playground.add_user(
            self.group, self.scenario.user, Group.Roles.ADMIN
        )
        self.client.force_login(self.scenario.user)
        response = self.client.post(
            self.group.get_absolute_url(), data={"username": "does-not-exist"}
        )
        tree = self.assertResponseHTML(response)
        self.assertEphemeral(tree, False)
        self.assertFormErrors(tree, True)

    def test_workspace_scoping(self) -> None:
        ws_group = self.playground.create_group(
            name=self.group.name,
            scope=self.group.scope,
            workspace=self.scenario.workspace,
            users=[self.scenario.user],
        )
        for group in (self.group, ws_group):
            with self.subTest(group=repr(group)):
                response = self.assertGetSucceeds(group)
                self.assertEqual(response.context["group"].pk, group.pk)


class GroupDetailViewMemberPaginationTests(ViewTestMixin, TestCase):
    """Tests pagination for :py:class:`GroupDetailView`."""

    scenario = scenarios.DefaultContext()
    group: ClassVar[Group]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.group = cls.playground.create_group("testgroup")
        # Add enough users to trigger pagination
        for idx in range(20):
            user = cls.playground.create_user(f"user{idx:02d}")
            GroupMembership.objects.create(
                group=cls.group,
                user=user,
                # Admins are the odd ones
                role=Group.Roles.ADMIN if idx % 2 == 1 else Group.Roles.MEMBER,
            )

    def _get(self, **kwargs: Any) -> TestResponseType:
        """Fetch the page."""
        self.client.force_login(self.scenario.user)
        response = self.client.get(
            self.group.get_absolute_url() + "?" + urlencode(kwargs)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        return response

    def assertGroupTable(
        self,
        tree: lxml.objectify.ObjectifiedElement,
        expected: list[tuple[str, Group.Roles]],
    ) -> None:
        """Check the contents of the group members table."""
        members = self.assertHasElement(tree, "//table[@id='members']")
        actual: list[tuple[str, Group.Roles]] = []
        for tr in members.tbody.tr:
            username = self.get_node_text_normalized(tr.td[0])
            role = Group.Roles(self.get_node_text_normalized(tr.td[2]))
            actual.append((username, role))

        self.assertEqual(actual, expected)

    def test_default(self) -> None:
        """Test default view with no page or ordering selected."""
        response = self._get()
        tree = self.assertResponseHTML(response)
        ctx = response.context
        self.assertEqual(ctx["asc"], "1")
        self.assertIsNone(ctx["order"])
        self.assertGroupTable(
            tree,
            [
                ("user01", Group.Roles.ADMIN),
                ("user03", Group.Roles.ADMIN),
                ("user05", Group.Roles.ADMIN),
                ("user07", Group.Roles.ADMIN),
                ("user09", Group.Roles.ADMIN),
                ("user11", Group.Roles.ADMIN),
                ("user13", Group.Roles.ADMIN),
                ("user15", Group.Roles.ADMIN),
                ("user17", Group.Roles.ADMIN),
                ("user19", Group.Roles.ADMIN),
            ],
        )

    def test_page(self) -> None:
        """Test selecting the page number."""
        response = self._get(page=2)
        tree = self.assertResponseHTML(response)
        ctx = response.context
        self.assertEqual(ctx["asc"], "1")
        self.assertIsNone(ctx["order"])
        self.assertGroupTable(
            tree,
            [
                ("user00", Group.Roles.MEMBER),
                ("user02", Group.Roles.MEMBER),
                ("user04", Group.Roles.MEMBER),
                ("user06", Group.Roles.MEMBER),
                ("user08", Group.Roles.MEMBER),
                ("user10", Group.Roles.MEMBER),
                ("user12", Group.Roles.MEMBER),
                ("user14", Group.Roles.MEMBER),
                ("user16", Group.Roles.MEMBER),
                ("user18", Group.Roles.MEMBER),
            ],
        )

    def test_order_username(self) -> None:
        """Test ordering by username."""
        response = self._get(order="name")
        tree = self.assertResponseHTML(response)
        ctx = response.context
        self.assertEqual(ctx["order"], "name")
        self.assertGroupTable(
            tree,
            [
                ("user00", Group.Roles.MEMBER),
                ("user01", Group.Roles.ADMIN),
                ("user02", Group.Roles.MEMBER),
                ("user03", Group.Roles.ADMIN),
                ("user04", Group.Roles.MEMBER),
                ("user05", Group.Roles.ADMIN),
                ("user06", Group.Roles.MEMBER),
                ("user07", Group.Roles.ADMIN),
                ("user08", Group.Roles.MEMBER),
                ("user09", Group.Roles.ADMIN),
            ],
        )

        response = self._get(order="name.desc")
        tree = self.assertResponseHTML(response)
        ctx = response.context
        self.assertEqual(ctx["order"], "name.desc")
        self.assertGroupTable(
            tree,
            [
                ("user19", Group.Roles.ADMIN),
                ("user18", Group.Roles.MEMBER),
                ("user17", Group.Roles.ADMIN),
                ("user16", Group.Roles.MEMBER),
                ("user15", Group.Roles.ADMIN),
                ("user14", Group.Roles.MEMBER),
                ("user13", Group.Roles.ADMIN),
                ("user12", Group.Roles.MEMBER),
                ("user11", Group.Roles.ADMIN),
                ("user10", Group.Roles.MEMBER),
            ],
        )

    def test_order_role(self) -> None:
        """Test ordering by role."""
        response = self._get(order="role")
        tree = self.assertResponseHTML(response)
        ctx = response.context
        self.assertEqual(ctx["order"], "role")
        self.assertGroupTable(
            tree,
            [
                ('user01', Group.Roles.ADMIN),
                ('user03', Group.Roles.ADMIN),
                ('user05', Group.Roles.ADMIN),
                ('user07', Group.Roles.ADMIN),
                ('user09', Group.Roles.ADMIN),
                ('user11', Group.Roles.ADMIN),
                ('user13', Group.Roles.ADMIN),
                ('user15', Group.Roles.ADMIN),
                ('user17', Group.Roles.ADMIN),
                ('user19', Group.Roles.ADMIN),
            ],
        )

        response = self._get(order="role.desc")
        tree = self.assertResponseHTML(response)
        ctx = response.context
        self.assertEqual(ctx["order"], "role.desc")
        self.assertGroupTable(
            tree,
            [
                ('user00', Group.Roles.MEMBER),
                ('user02', Group.Roles.MEMBER),
                ('user04', Group.Roles.MEMBER),
                ('user06', Group.Roles.MEMBER),
                ('user08', Group.Roles.MEMBER),
                ('user10', Group.Roles.MEMBER),
                ('user12', Group.Roles.MEMBER),
                ('user14', Group.Roles.MEMBER),
                ('user16', Group.Roles.MEMBER),
                ('user18', Group.Roles.MEMBER),
            ],
        )


class MembershipUpdateViewTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`MembershipUpdateView`."""

    scenario = scenarios.DefaultContext()
    group: ClassVar[Group]
    other_user: ClassVar[User]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.group = cls.playground.create_group("testgroup")
        cls.playground.add_user(cls.group, cls.scenario.user, Group.Roles.ADMIN)
        cls.other_user = cls.playground.create_user("other")
        cls.playground.add_user(cls.group, cls.other_user)

    def _get(self, user: User, group: Group | None = None) -> TestResponseType:
        """GET to the view for the given user."""
        if group is None:
            group = self.group
        self.client.force_login(self.scenario.user)
        gm = GroupMembership.objects.get(group=group, user=user)
        return self.client.get(gm.get_absolute_url_update())

    def _post(self, user: User, data: dict[str, Any]) -> TestResponseType:
        """POST to the view for the given user."""
        self.client.force_login(self.scenario.user)
        gm = GroupMembership.objects.get(group=self.group, user=user)
        url = gm.get_absolute_url_update()
        return self.client.post(url, data=data)

    def assertRole(self, user: User, role: Group.Roles) -> None:
        """Ensure the user has the given role."""
        membership = GroupMembership.objects.get(user=user, group=self.group)
        self.assertEqual(membership.role, role)

    def test_permissions(self) -> None:
        """Test that the right permissions are enforced."""
        gm = GroupMembership.objects.get(
            group=self.group, user=self.scenario.user
        )
        url = gm.get_absolute_url_update()
        self.client.force_login(self.scenario.user)

        self.assertEnforcesPermission(
            self.group.can_manage,
            url,
            "get_context_data",
            method="get",
            error_code=status.HTTP_404_NOT_FOUND,
        )

        self.assertEnforcesPermission(
            self.group.can_manage,
            url,
            "form_valid",
            method="post",
            data={"role": "admin"},
            error_code=status.HTTP_404_NOT_FOUND,
        )

    def test_role_validation(self) -> None:
        """Test role validation."""
        response = self._post(
            self.scenario.user, data={"role": "does-not-exist"}
        )
        tree = self.assertResponseHTML(response)
        errorlist = self.assertHasElement(tree, "//ul[@class='errorlist']")
        self.assertEqual(len(errorlist.li), 1)
        self.assertTextContentEqual(
            errorlist.li,
            "Select a valid choice. does-not-exist"
            " is not one of the available choices.",
        )

    def test_intro_self(self) -> None:
        """Check page introduction for oneself."""
        response = self._get(self.scenario.user)
        tree = self.assertResponseHTML(response)
        intro = self.assertHasElement(tree, "//p[@id='intro']")
        self.assertTextContentEqual(
            intro,
            f"Update your role in group {self.group}.",
        )

    def test_intro_other(self) -> None:
        """Check page introduction for another user."""
        response = self._get(self.other_user)
        tree = self.assertResponseHTML(response)
        intro = self.assertHasElement(tree, "//p[@id='intro']")
        self.assertTextContentEqual(
            intro,
            f"Update role of user {self.other_user.get_full_name()}"
            f" <{self.other_user.username}>"
            f" in group {self.group}.",
        )

    def test_change_other(self) -> None:
        """Test changing the role of another user."""
        # Can set another person as admin
        response = self._post(self.other_user, data={"role": Group.Roles.ADMIN})
        self.assertRedirects(response, self.group.get_absolute_url())
        self.assertRole(self.other_user, Group.Roles.ADMIN)

        # Can revoke another ADMIN role
        response = self._post(
            self.other_user, data={"role": Group.Roles.MEMBER}
        )
        self.assertRedirects(response, self.group.get_absolute_url())
        self.assertRole(self.other_user, Group.Roles.MEMBER)

    def test_change_self(self) -> None:
        """Test changing the role of oneself."""
        # Can set oneself as admin, only if already admin
        response = self._post(
            self.scenario.user, data={"role": Group.Roles.ADMIN}
        )
        self.assertRedirects(response, self.group.get_absolute_url())
        self.assertRole(self.scenario.user, Group.Roles.ADMIN)

        # Can revoke one's own ADMIN role
        response = self._post(
            self.scenario.user, data={"role": Group.Roles.MEMBER}
        )
        self.assertRedirects(response, self.group.get_absolute_url())
        self.assertRole(self.scenario.user, Group.Roles.MEMBER)

        # This locks oneself out
        response = self._post(
            self.scenario.user, data={"role": Group.Roles.ADMIN}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_workspace_scoping(self) -> None:
        ws_group = self.playground.create_group(
            name=self.group.name,
            scope=self.group.scope,
            workspace=self.scenario.workspace,
        )
        self.playground.add_user(
            ws_group, self.scenario.user, Group.Roles.ADMIN
        )
        for group in (self.group, ws_group):
            with self.subTest(group=repr(group)):
                response = self._get(self.scenario.user, group)
                self.assertResponseHTML(response)
                self.assertEqual(response.context["object"].group.pk, group.pk)


class MembershipDeleteViewTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`MembershipDeleteView`."""

    scenario = scenarios.DefaultContext()
    group: ClassVar[Group]
    other_user: ClassVar[User]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.group = cls.playground.create_group("testgroup")
        cls.playground.add_user(cls.group, cls.scenario.user, Group.Roles.ADMIN)
        cls.other_user = cls.playground.create_user("other")
        cls.playground.add_user(cls.group, cls.other_user)

    def assertIsMember(self, user: User) -> None:
        """Ensure the user is a member of self.group."""
        self.assertTrue(
            GroupMembership.objects.filter(group=self.group, user=user).exists()
        )

    def assertIsNotMember(self, user: User) -> None:
        """Ensure the user is a member of self.group."""
        self.assertFalse(
            GroupMembership.objects.filter(group=self.group, user=user).exists()
        )

    def _get(self, user: User, group: Group | None = None) -> TestResponseType:
        """GET to the view for the given user."""
        if group is None:
            group = self.group
        self.client.force_login(self.scenario.user)
        gm = GroupMembership.objects.get(group=group, user=user)
        return self.client.get(gm.get_absolute_url_remove())

    def _post(self, user: User, group: Group | None = None) -> TestResponseType:
        """POST to the view for the given user."""
        if group is None:
            group = self.group
        self.client.force_login(self.scenario.user)
        gm = GroupMembership.objects.get(group=group, user=user)
        return self.client.post(gm.get_absolute_url_remove())

    def test_permissions(self) -> None:
        """Test that the right permissions are enforced."""
        gm = GroupMembership.objects.get(
            group=self.group, user=self.scenario.user
        )
        url = gm.get_absolute_url_remove()
        self.client.force_login(self.scenario.user)

        self.assertEnforcesPermission(
            self.group.can_manage,
            url,
            "get_context_data",
            method="get",
            error_code=status.HTTP_404_NOT_FOUND,
        )

        self.assertEnforcesPermission(
            self.group.can_manage,
            url,
            "form_valid",
            method="post",
            data={},
            error_code=status.HTTP_404_NOT_FOUND,
        )

    def test_intro_self(self) -> None:
        """Check page introduction for oneself."""
        response = self._get(self.scenario.user)
        tree = self.assertResponseHTML(response)
        intro = self.assertHasElement(tree, "//p[@id='intro']")
        self.assertTextContentEqual(
            intro,
            f"Are you sure you want to remove yourself"
            f" from group {self.group.name} in scope debusine?",
        )

    def test_intro_other(self) -> None:
        """Check page introduction for another user."""
        response = self._get(self.other_user)
        tree = self.assertResponseHTML(response)
        intro = self.assertHasElement(tree, "//p[@id='intro']")
        self.assertTextContentEqual(
            intro,
            "Are you sure you want to remove user "
            f" {self.other_user.get_full_name()} "
            f" <{self.other_user.username}>"
            f" from group {self.group.name} in scope debusine?",
        )

    def test_remove_other(self) -> None:
        """Test removing another user."""
        response = self._post(self.other_user)
        self.assertRedirects(response, self.group.get_absolute_url())
        self.assertIsMember(self.scenario.user)
        self.assertIsNotMember(self.other_user)

    def test_remove_self(self) -> None:
        """Test removing oneself."""
        response = self._post(self.scenario.user)
        self.assertRedirects(response, self.group.get_absolute_url())
        self.assertIsNotMember(self.scenario.user)
        self.assertIsMember(self.other_user)

    def test_workspace_scoping(self) -> None:
        ws_group = self.playground.create_group(
            name=self.group.name,
            scope=self.group.scope,
            workspace=self.scenario.workspace,
        )
        self.playground.add_user(
            ws_group, self.scenario.user, Group.Roles.ADMIN
        )
        for group in (self.group, ws_group):
            with self.subTest(group=repr(group)):
                response = self._get(self.scenario.user, group)
                self.assertResponseHTML(response)
                self.assertEqual(response.context["object"].group.pk, group.pk)

                response = self._post(self.scenario.user, group)
                self.assertRedirects(response, group.get_absolute_url())


class GroupAuditLogViewTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`GroupAuditLogView`."""

    scenario = scenarios.DefaultContext()
    group: ClassVar[Group]
    other_user: ClassVar[User]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.other_user = cls.playground.create_user("other")
        cls.group = cls.playground.create_group("testgroup")
        changes = auth.GroupAuditLogMemberRemoved(user="…")
        for n in range(0, 6, 2):
            changes.user = f"{n}"
            GroupAuditLog.objects.create(
                group=cls.group,
                actor=cls.scenario.user,
                changes=changes.model_dump(),
            )
            changes.user = f"{n + 1}"
            GroupAuditLog.objects.create(
                group=cls.group,
                actor=cls.other_user,
                changes=changes.model_dump(),
            )

    def assertGetSucceeds(self, group: Group | None = None) -> TestResponseType:
        """GET the page and ensure the request succeeds."""
        if group is None:
            group = self.group
        self.client.force_login(self.scenario.user)
        response = self.client.get(group.get_absolute_url_audit_log())
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        return response

    def assertHasRows(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> list[int]:
        """Ensure the table has rows and return the list of messages."""
        res: list[int] = []
        table = self.assertHasElement(tree, "//table[@id='audit-log']")
        for tr in table.tbody.tr:
            message = self.get_node_text_normalized(tr.td[2])
            res.append(int(message.removeprefix("removed ")))
        self.assertTrue(res)
        return res

    def test_permissions(self) -> None:
        """Test that the right permissions are enforced."""
        url = self.group.get_absolute_url_audit_log()
        self.client.force_login(self.scenario.user)

        self.assertEnforcesPermission(
            self.group.can_display,
            url,
            "get_context_data",
            method="get",
            error_code=status.HTTP_404_NOT_FOUND,
        )

    def test_get_as_anonymous(self) -> None:
        """Test that the right permissions are enforced."""
        response = self.client.get(self.group.get_absolute_url_audit_log())
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_get_as_nonmember(self) -> None:
        """View the page as a group member."""
        response = self.assertGetSucceeds()
        tree = self.assertResponseHTML(response)
        rows = self.assertHasRows(tree)
        self.assertEqual(rows, [5, 4, 3, 2, 1, 0])

    def test_get_as_member(self) -> None:
        """View the page as a group member."""
        self.playground.add_user(self.group, self.scenario.user)
        response = self.assertGetSucceeds()
        tree = self.assertResponseHTML(response)
        rows = self.assertHasRows(tree)
        self.assertEqual(rows, [5, 4, 3, 2, 1, 0])

    def test_get_as_admin(self) -> None:
        """View the page as a group member."""
        self.playground.add_user(
            self.group, self.scenario.user, Group.Roles.ADMIN
        )
        response = self.assertGetSucceeds()
        tree = self.assertResponseHTML(response)
        rows = self.assertHasRows(tree)
        self.assertEqual(rows, [5, 4, 3, 2, 1, 0])

    def test_workspace_scoping(self) -> None:
        ws_group = self.playground.create_group(
            name=self.group.name,
            scope=self.group.scope,
            workspace=self.scenario.workspace,
            users=[self.scenario.user],
        )
        for group in (self.group, ws_group):
            with self.subTest(group=repr(group)):
                response = self.assertGetSucceeds(group)
                self.assertEqual(response.context["group"].pk, group.pk)
